# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Collection of all constants used in pySpeos."""

import os

DEFAULT_HOST: str = "localhost"
"""Default host used by Speos RPC server and client """
DEFAULT_PORT: str = "50098"
"""Default port used by Speos RPC server and client """
DEFAULT_VERSION: str = "251"
"""Latest supported Speos version of the current PySpeos Package"""
MAX_SERVER_MESSAGE_LENGTH: int = int(os.environ.get("SPEOS_MAX_MESSAGE_LENGTH", 256 * 1024**2))
"""Maximum message length value accepted by the Speos RPC server,
By default, value stored in environment variable SPEOS_MAX_MESSAGE_LENGTH or 268 435 456.
"""
MAX_CLIENT_MESSAGE_SIZE: int = int(4194304)
"""Maximum message Size accepted by grpc channel,
By default, 4194304.
"""



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a way to gather Speos features."""

from __future__ import annotations

from pathlib import Path
import re
from typing import TYPE_CHECKING, List, Mapping, Optional, Union
import uuid

from google.protobuf.internal.containers import RepeatedScalarFieldContainer
import numpy as np

import ansys.speos.core.body as body
import ansys.speos.core.face as face
from ansys.speos.core.generic.general_methods import graphics_required
from ansys.speos.core.kernel.body import BodyLink
from ansys.speos.core.kernel.face import FaceLink
from ansys.speos.core.kernel.part import ProtoPart
from ansys.speos.core.kernel.scene import ProtoScene
import ansys.speos.core.opt_prop as opt_prop
import ansys.speos.core.part as part
import ansys.speos.core.proto_message_utils as proto_message_utils
from ansys.speos.core.sensor import (
    SensorCamera,
    SensorIrradiance,
    SensorRadiance,
)
from ansys.speos.core.simulation import (
    SimulationDirect,
    SimulationInteractive,
    SimulationInverse,
)
from ansys.speos.core.source import (
    SourceLuminaire,
    SourceRayFile,
    SourceSurface,
)
from ansys.speos.core.speos import Speos

try:
    from ansys.speos.core.generic.general_methods import run_if_graphics_required

    run_if_graphics_required(warning=True)
except ImportError as err:  # pragma: no cover
    raise err

if TYPE_CHECKING:  # pragma: no cover
    import pyvista as pv

    from ansys.tools.visualization_interface import Plotter


class Project:
    """A project describes all Speos features.

    This includes optical properties, sources, sensors, simulations that user can fill in.
    Project provides functions to create new feature, find a feature.
    It can be created from empty or loaded from a specific file.

    Parameters
    ----------
    speos : ansys.speos.core.speos.Speos
        Speos session (connected to gRPC server).
    path : str
        The project will be loaded from this speos file.
        By default, ``""``, means create from empty.

    Attributes
    ----------
    scene_link : ansys.speos.core.kernel.scene.SceneLink
        Link object for the scene in database.
    """

    def __init__(self, speos: Speos, path: str = ""):
        self.client = speos.client
        """Speos instance client."""
        self.scene_link = speos.client.scenes().create()
        """Link object for the scene in database."""
        self._features = []
        if len(path):
            self.scene_link.load_file(path)
            self._fill_features()

    # def list(self):
    #    """Return all feature key as a tree.
    #
    #    Can be used to list all features- Not yet implemented.
    #    """
    #    pass

    def create_optical_property(
        self,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
    ) -> opt_prop.OptProp:
        """Create a new Optical Property feature.

        Parameters
        ----------
        name : str
            Name of the feature.
        description : str
            Description of the feature.
            By default, ``""``.
        metadata : Optional[Mapping[str, str]]
            Metadata of the feature.
            By default, ``{}``.

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            OptProp feature.
        """
        existing_features = self.find(name=name)
        if len(existing_features) != 0:
            msg = "Feature {}: {} has a conflict name with an existing feature.".format(
                opt_prop.OptProp, name
            )
            raise ValueError(msg)

        if metadata is None:
            metadata = {}
        feature = opt_prop.OptProp(
            project=self, name=name, description=description, metadata=metadata
        )
        self._features.append(feature)
        return feature

    def create_source(
        self,
        name: str,
        description: str = "",
        feature_type: type = SourceSurface,
        metadata: Optional[Mapping[str, str]] = None,
    ) -> Union[SourceSurface, SourceRayFile, SourceLuminaire]:
        """Create a new Source feature.

        Parameters
        ----------
        name : str
            Name of the feature.
        description : str
            Description of the feature.
            By default, ``""``.
        feature_type: type
            Source type to be created.
            By default, ``ansys.speos.core.source.SourceSurface``.
            Allowed types:
            Union[ansys.speos.core.source.SourceSurface, ansys.speos.core.source.SourceRayFile, \
            ansys.speos.core.source.SourceLuminaire].
        metadata : Optional[Mapping[str, str]]
            Metadata of the feature.
            By default, ``{}``.

        Returns
        -------
        Union[ansys.speos.core.source.SourceSurface,ansys.speos.core.source.SourceRayFile,\
        ansys.speos.core.source.SourceLuminaire]
            Source class instance.
        """
        if metadata is None:
            metadata = {}

        existing_features = self.find(name=name)
        if len(existing_features) != 0:
            msg = "Feature {}: {} has a conflict name with an existing feature.".format(
                feature_type, name
            )
            raise ValueError(msg)
        feature = None
        if feature_type == SourceSurface:
            feature = SourceSurface(
                project=self,
                name=name,
                description=description,
                metadata=metadata,
            )
        elif feature_type == SourceRayFile:
            feature = SourceRayFile(
                project=self,
                name=name,
                description=description,
                metadata=metadata,
            )
        elif feature_type == SourceLuminaire:
            feature = SourceLuminaire(
                project=self,
                name=name,
                description=description,
                metadata=metadata,
            )
        else:
            msg = "Requested feature {} does not exist in supported list {}".format(
                feature_type, [SourceSurface, SourceLuminaire, SourceRayFile]
            )
            raise TypeError(msg)
        self._features.append(feature)
        return feature

    def create_simulation(
        self,
        name: str,
        description: str = "",
        feature_type: type = SimulationDirect,
        metadata: Optional[Mapping[str, str]] = None,
    ) -> Union[SimulationDirect, SimulationInteractive, SimulationInverse]:
        """Create a new Simulation feature.

        Parameters
        ----------
        name : str
            Name of the feature.
        description : str
            Description of the feature.
            By default, ``""``.
        feature_type: type
            Simulation type to be created.
            By default, ``ansys.speos.core.simulation.SimulationDirect``.
            Allowed types: Union[ansys.speos.core.simulation.SimulationDirect, \
            ansys.speos.core.simulation.SimulationInteractive, \
            ansys.speos.core.simulation.SimulationInverse].
        metadata : Optional[Mapping[str, str]]
            Metadata of the feature.
            By default, ``{}``.

        Returns
        -------
        Union[ansys.speos.core.simulation.SimulationDirect,\
        ansys.speos.core.simulation.SimulationInteractive,\
        ansys.speos.core.simulation.SimulationInverse]
            Simulation class instance
        """
        if metadata is None:
            metadata = {}

        existing_features = self.find(name=name)
        if len(existing_features) != 0:
            msg = "Feature {}: {} has a conflict name with an existing feature.".format(
                feature_type, name
            )
            raise ValueError(msg)
        feature = None
        if feature_type == SimulationDirect:
            feature = SimulationDirect(
                project=self,
                name=name,
                description=description,
                metadata=metadata,
            )
        elif feature_type == SimulationInverse:
            feature = SimulationInverse(
                project=self,
                name=name,
                description=description,
                metadata=metadata,
            )
        elif feature_type == SimulationInteractive:
            feature = SimulationInteractive(
                project=self,
                name=name,
                description=description,
                metadata=metadata,
            )
        else:
            msg = "Requested feature {} does not exist in supported list {}".format(
                feature_type,
                [
                    SimulationDirect,
                    SimulationInverse,
                    SimulationInteractive,
                ],
            )
            raise TypeError(msg)
        self._features.append(feature)
        return feature

    def create_sensor(
        self,
        name: str,
        description: str = "",
        feature_type: type = SensorIrradiance,
        metadata: Optional[Mapping[str, str]] = None,
    ) -> Union[SensorCamera, SensorRadiance, SensorIrradiance]:
        """Create a new Sensor feature.

        Parameters
        ----------
        name : str
            Name of the feature.
        description : str
            Description of the feature.
            By default, ``""``.
        feature_type: type
            Sensor type to be created.
            By default, ``ansys.speos.core.sensor.SensorIrradiance``.
            Allowed types: Union[ansys.speos.core.sensor.SensorCamera,\
            ansys.speos.core.sensor.SensorRadiance, \
            ansys.speos.core.sensor.SensorIrradiance].
        metadata : Optional[Mapping[str, str]]
            Metadata of the feature.
            By default, ``{}``.

        Returns
        -------
        Union[ansys.speos.core.sensor.SensorCamera,\
        ansys.speos.core.sensor.SensorRadiance, ansys.speos.core.sensor.SensorIrradiance]
            Sensor class instance.
        """
        if metadata is None:
            metadata = {}

        existing_features = self.find(name=name)
        if len(existing_features) != 0:
            msg = "Feature {}: {} has a conflict name with an existing feature.".format(
                feature_type, name
            )
            raise ValueError(msg)
        feature = None
        if feature_type == SensorIrradiance:
            feature = SensorIrradiance(
                project=self,
                name=name,
                description=description,
                metadata=metadata,
            )
        elif feature_type == SensorRadiance:
            feature = SensorRadiance(
                project=self,
                name=name,
                description=description,
                metadata=metadata,
            )
        elif feature_type == SensorCamera:
            feature = SensorCamera(
                project=self,
                name=name,
                description=description,
                metadata=metadata,
            )
        else:
            msg = "Requested feature {} does not exist in supported list {}".format(
                feature_type, [SensorIrradiance, SensorRadiance, SensorCamera]
            )
            raise TypeError(msg)
        self._features.append(feature)
        return feature

    def create_root_part(
        self,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
    ) -> part.Part:
        """Create the project root part feature.

        If a root part is already created in the project, it is returned.

        Parameters
        ----------
        description : str
            Description of the feature.
            By default, ``""``.
        metadata : Optional[Mapping[str, str]]
            Metadata of the feature.
            By default, ``{}``.

        Returns
        -------
        ansys.speos.core.part.Part
            Part feature.
        """
        if metadata is None:
            metadata = {}

        name = "RootPart"
        existing_rp = self.find(name="", feature_type=part.Part)
        if existing_rp:
            return existing_rp[0]

        feature = part.Part(project=self, name=name, description=description, metadata=metadata)
        self._features.append(feature)
        return feature

    def find(
        self,
        name: str,
        name_regex: bool = False,
        feature_type: Optional[type] = None,
    ) -> List[
        Union[
            opt_prop.OptProp,
            SourceSurface,
            SourceLuminaire,
            SourceRayFile,
            SensorIrradiance,
            SensorRadiance,
            SensorCamera,
            SimulationDirect,
            SimulationInverse,
            SimulationInteractive,
            part.Part,
            body.Body,
            face.Face,
            part.Part.SubPart,
        ]
    ]:
        """Find feature(s) by name (possibility to use regex) and by feature type.

        Parameters
        ----------
        name : str
            Name of the feature.
        name_regex : bool
            Allows to use regex for name parameter.
            By default, ``False``, means that regex is not used for name parameter.
        feature_type : type
            Type of the wanted features.
            Mandatory to fill for geometry features.
            By default, ``None``, means that all features will be considered
            (except geometry features).

        Returns
        -------
        List[Union[ansys.speos.core.opt_prop.OptProp, ansys.speos.core.source.Surface, \
        ansys.speos.core.source.RayFile, ansys.speos.core.source.Luminaire, \
        ansys.speos.core.sensor.Camera, \
        ansys.speos.core.sensor.Radiance, ansys.speos.core.sensor.Irradiance, \
        ansys.speos.core.simulation.Direct, ansys.speos.core.simulation.Interactive, \
        ansys.speos.core.simulation.Inverse, ansys.speos.core.part.Part, \
        ansys.speos.core.body.Body, \
        ansys.speos.core.face.Face, ansys.speos.core.part.Part.SubPart]]
            Found features.

        Examples
        --------
        >>> # From name only
        >>> find(name="Camera.1")
        >>> # Specify feature type
        >>> find(name="Camera.1", feature_type=ansys.speos.core.sensor.SensorCamera)
        >>> # Using regex
        >>> find(
        >>>     name="Camera.*",
        >>>     name_regex=True,
        >>>     feature_type=ansys.speos.core.sensor.SensorCamera,
        >>> )
        Here some examples when looking for a geometry feature:
        (always precise feature_type)

        >>> # Root part
        >>> find(name="", feature_type=ansys.speos.core.part.Part)
        >>> # Body in root part
        >>> find(name="BodyName", feature_type=ansys.speos.core.body.Body)
        >>> # Face from body in root part
        >>> find(name="BodyName/FaceName", feature_type=ansys.speos.core.face.Face)
        >>> # Sub part in root part
        >>> find(name="SubPartName", feature_type=ansys.speos.core.part.Part.SubPart)
        >>> # Face in a body from sub part in root part :
        >>> find(name="SubPartName/BodyName/FaceName", feature_type=ansys.speos.core.face.Face)
        >>> # Regex can be use at each level separated by "/"
        >>> find(name="Body.*/Face.*", name_regex=True, feature_type=ansys.speos.core.face.Face)
        >>> # All faces of a specific body
        >>> find(name="BodyName/.*", name_regex=True, feature_type=ansys.speos.core.face.Face)
        >>> # All geometry features at first level (whatever their type: body, face, sub part)
        >>> find(name=".*", name_regex=True, feature_type=ansys.speos.core.part.Part)
        """
        orig_feature_type = None
        if (
            feature_type == part.Part
            or feature_type == part.Part.SubPart
            or feature_type == body.Body
            or feature_type == face.Face
        ):
            if feature_type != part.Part:
                orig_feature_type = feature_type
                feature_type = part.Part
            if name == "":
                name = "RootPart"
            else:
                name = "RootPart/" + name

        orig_name = name
        idx = name.find("/")
        if idx != -1:
            name = name[0:idx]

        if name_regex:
            p = re.compile(name)

        found_features = []
        if feature_type is None:
            if name_regex:
                found_features.extend([x for x in self._features if p.match(x._name)])
            else:
                found_features.extend([x for x in self._features if x._name == name])
        else:
            if name_regex:
                found_features.extend(
                    [
                        x
                        for x in self._features
                        if (
                            isinstance(x, feature_type)
                            or (isinstance(x._type, feature_type) if hasattr(x, "_type") else False)
                        )
                        and p.match(x._name)
                    ]
                )
            else:
                found_features.extend(
                    [
                        x
                        for x in self._features
                        if (
                            isinstance(x, feature_type)
                            or (isinstance(x._type, feature_type) if hasattr(x, "_type") else False)
                        )
                        and x._name == name
                    ]
                )

        if found_features and idx != -1:
            tmp = [
                f.find(
                    name=orig_name[idx + 1 :],
                    name_regex=name_regex,
                    feature_type=orig_feature_type,
                )
                for f in found_features
            ]

            found_features.clear()
            for feats in tmp:
                found_features.extend(feats)

        return found_features

    # def action(self, name: str):
    #    """Act on feature: update, hide/show, copy, ... - Not yet implemented"""
    #    pass

    # def save(self):
    #     """Save class state in file given at construction - Not yet implemented"""
    #     pass

    def delete(self) -> Project:
        """Delete project: erase scene data.

        Delete all features contained in the project.

        Returns
        -------
        ansys.speos.core.project.Project
            Project feature.
        """
        # Erase the scene
        if self.scene_link is not None:
            self.scene_link.set(data=ProtoScene())

        # Delete each feature that was created
        for f in self._features:
            f.delete()
        self._features.clear()

        return self

    def _to_dict(self) -> dict:
        # Replace all guids by content of objects in the dict
        output_dict = proto_message_utils._replace_guids(
            speos_client=self.client,
            message=self.scene_link.get(),
            ignore_simple_key="part_guid",
        )

        # For each feature, replace properties by putting them at correct place
        for k, v in output_dict.items():
            if type(v) is list:
                for inside_dict in v:
                    if k == "simulations":
                        sim_feat = self.find(
                            name=inside_dict["name"],
                            feature_type=SimulationDirect,
                        )
                        if len(sim_feat) == 0:
                            sim_feat = self.find(
                                name=inside_dict["name"],
                                feature_type=SimulationInverse,
                            )
                        if len(sim_feat) == 0:
                            sim_feat = self.find(
                                name=inside_dict["name"],
                                feature_type=SimulationInteractive,
                            )
                        sim_feat = sim_feat[0]
                        if sim_feat.job_link is None:
                            inside_dict["simulation_properties"] = (
                                proto_message_utils._replace_guids(
                                    speos_client=self.client,
                                    message=sim_feat._job,
                                    ignore_simple_key="scene_guid",
                                )
                            )
                        else:
                            inside_dict["simulation_properties"] = (
                                proto_message_utils._replace_guids(
                                    speos_client=self.client,
                                    message=sim_feat.job_link.get(),
                                    ignore_simple_key="scene_guid",
                                )
                            )

                    proto_message_utils._replace_properties(inside_dict)
        return output_dict

    def get(self) -> dict:
        """Get dictionary corresponding to the project - read only."""
        return self._to_dict()

    def find_key(self, key: str) -> List[tuple[str, dict]]:
        """Get values corresponding to the key in project dictionary - read only.

        Parameters
        ----------
        key : str
            Key to search in the project dictionary.

        Returns
        -------
        List[tuple[str, dict]]
            List of matching objects containing for each its x_path and its value.
        """
        return proto_message_utils._finder_by_key(dict_var=self._to_dict(), key=key)

    def __str__(self):
        """Return the string representation of the project's scene."""
        return proto_message_utils.dict_to_str(dict=self._to_dict())

    def _fill_bodies(
        self,
        body_guids: List[str],
        feat_host: Union[part.Part, part.Part.SubPart],
    ):
        """Fill part of sub part features from a list of body guids."""
        for b_link in self.client.get_items(keys=body_guids, item_type=BodyLink):
            b_data = b_link.get()
            b_feat = feat_host.create_body(name=b_data.name)
            b_feat.body_link = b_link
            b_feat._body = b_data  # instead of b_feat.reset() - this avoid a useless read in server
            f_links = self.client.get_items(keys=b_data.face_guids, item_type=FaceLink)
            for f_link in f_links:
                f_data = f_link.get()
                f_feat = b_feat.create_face(name=f_data.name)
                f_feat.face_link = f_link
                f_feat._face = (
                    f_data  # instead of f_feat.reset() - this avoid a useless read in server
                )

    def _add_unique_ids(self):
        scene_data = self.scene_link.get()

        root_part_link = self.client[scene_data.part_guid]
        root_part = root_part_link.get()
        update_rp = False
        for sub_part in root_part.parts:
            if sub_part.description.startswith("UniqueId_") is False:
                sub_part.description = "UniqueId_" + str(uuid.uuid4())
                update_rp = True
        if update_rp:
            root_part_link.set(data=root_part)

        for mat_inst in scene_data.materials:
            if mat_inst.metadata["UniqueId"] == "":
                mat_inst.metadata["UniqueId"] = str(uuid.uuid4())

        for src_inst in scene_data.sources:
            if src_inst.metadata["UniqueId"] == "":
                src_inst.metadata["UniqueId"] = str(uuid.uuid4())

        for ssr_inst in scene_data.sensors:
            if ssr_inst.metadata["UniqueId"] == "":
                ssr_inst.metadata["UniqueId"] = str(uuid.uuid4())

        for sim_inst in scene_data.simulations:
            if sim_inst.metadata["UniqueId"] == "":
                sim_inst.metadata["UniqueId"] = str(uuid.uuid4())

        self.scene_link.set(data=scene_data)

    def _fill_features(self):
        """Fill project features from a scene."""
        self._add_unique_ids()

        scene_data = self.scene_link.get()

        root_part_link = self.client[scene_data.part_guid]
        root_part_data = root_part_link.get()
        root_part_feats = self.find(name="", feature_type=part.Part)
        root_part_feat = None
        if not root_part_feats:
            root_part_feat = self.create_root_part()
            root_part_data.name = "RootPart"
            root_part_link.set(root_part_data)
            self._fill_bodies(body_guids=root_part_data.body_guids, feat_host=root_part_feat)
        else:
            root_part_feat = root_part_feats[0]

        root_part_feat.part_link = root_part_link
        root_part_feat._part = root_part_data
        # instead of root_part_feat.reset() - this avoid a useless read in server

        for sp in root_part_data.parts:
            sp_feat = root_part_feat.create_sub_part(name=sp.name, description=sp.description)
            if sp.description.startswith("UniqueId_"):
                idx = sp.description.find("_")
                sp_feat._unique_id = sp.description[idx + 1 :]
            sp_feat.part_link = self.client[sp.part_guid]
            part_data = sp_feat.part_link.get()
            sp_feat._part_instance = sp
            sp_feat._part = (
                part_data  # instead of sp_feat.reset() - this avoid a useless read in server
            )
            self._fill_bodies(body_guids=part_data.body_guids, feat_host=sp_feat)

        for mat_inst in scene_data.materials:
            if len(self.find(name=mat_inst.name)) == 0:
                op_feature = self.create_optical_property(name=mat_inst.name)
                op_feature._fill(mat_inst=mat_inst)

        for src_inst in scene_data.sources:
            src_feat = None
            if src_inst.HasField("rayfile_properties"):
                src_feat = SourceRayFile(
                    project=self,
                    name=src_inst.name,
                    source_instance=src_inst,
                    default_values=False,
                )
            elif src_inst.HasField("luminaire_properties"):
                src_feat = SourceLuminaire(
                    project=self,
                    name=src_inst.name,
                    source_instance=src_inst,
                    default_values=False,
                )
            elif src_inst.HasField("surface_properties"):
                src_feat = SourceSurface(
                    project=self,
                    name=src_inst.name,
                    source_instance=src_inst,
                    default_values=False,
                )
            if src_feat is not None:
                self._features.append(src_feat)

        for ssr_inst in scene_data.sensors:
            ssr_feat = None
            if ssr_inst.HasField("irradiance_properties"):
                ssr_feat = SensorIrradiance(
                    project=self,
                    name=ssr_inst.name,
                    sensor_instance=ssr_inst,
                    default_values=False,
                )
            elif ssr_inst.HasField("radiance_properties"):
                ssr_feat = SensorRadiance(
                    project=self,
                    name=ssr_inst.name,
                    sensor_instance=ssr_inst,
                    default_values=False,
                )
            elif ssr_inst.HasField("camera_properties"):
                ssr_feat = SensorCamera(
                    project=self,
                    name=ssr_inst.name,
                    sensor_instance=ssr_inst,
                    default_values=False,
                )
            self._features.append(ssr_feat)

        for sim_inst in scene_data.simulations:
            sim_feat = None
            simulation_template_link = self.client[sim_inst.simulation_guid].get()
            if simulation_template_link.HasField("direct_mc_simulation_template"):
                sim_feat = SimulationDirect(
                    project=self,
                    name=sim_inst.name,
                    simulation_instance=sim_inst,
                    default_values=False,
                )
            elif simulation_template_link.HasField("inverse_mc_simulation_template"):
                sim_feat = SimulationInverse(
                    project=self,
                    name=sim_inst.name,
                    simulation_instance=sim_inst,
                    default_values=False,
                )
            elif simulation_template_link.HasField("interactive_simulation_template"):
                sim_feat = SimulationInteractive(
                    project=self,
                    name=sim_inst.name,
                    simulation_instance=sim_inst,
                    default_values=False,
                )
            self._features.append(sim_feat)

    def __extract_part_mesh_info(
        self,
        part_data: ProtoPart,
        part_coordinate_info: RepeatedScalarFieldContainer = None,
    ) -> pv.PolyData:
        """Extract mesh data info from a part.

        Parameters
        ----------
        speos_client : ansys.speos.core.kernel.client.SpeosClient
            The Speos instance client.
        part_data: ansys.api.speos.part.v1.part_pb2.Part
            Part from scene.
        part_coordinate_info: RepeatedScalarFieldContainer
            message contains part coordinate info: origin, x_vector, y_vector, z_vector

        Returns
        -------
        pv.PolyData
            mesh data extracted.
        """
        import pyvista as pv

        def local2absolute(local_vertice: np.ndarray, coordinates) -> np.ndarray:
            """Convert local coordinate to global coordinate.

            Parameters
            ----------
            local_vertice: np.ndarray
                numpy array includes x, y, z info.

            Returns
            -------
            np.ndarray
                numpy array includes x, y, z info

            """
            global_origin = np.array(coordinates[:3])
            global_x = np.array(coordinates[3:6]) * local_vertice[0]
            global_y = np.array(coordinates[6:9]) * local_vertice[1]
            global_z = np.array(coordinates[9:]) * local_vertice[2]
            return global_origin + global_x + global_y + global_z

        part_coordinate = [
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            1.0,
        ]
        if part_coordinate_info is not None:
            part_coordinate = part_coordinate_info
        part_mesh_info = None
        for body_idx, body_guid in enumerate(part_data.body_guids):
            body_item_data = self.client[body_guid].get()
            for face_idx, face_guid in enumerate(body_item_data.face_guids):
                face_item_data = self.client[face_guid].get()
                vertices = np.array(face_item_data.vertices)
                facets = np.array(face_item_data.facets)
                vertices = vertices.reshape(-1, 3)
                vertices = np.array(
                    [local2absolute(vertice, part_coordinate) for vertice in vertices]
                )
                facets = facets.reshape(-1, 3)
                temp = np.full(facets.shape[0], 3)
                temp = np.vstack(temp)
                facets = np.hstack((temp, facets))
                face_mesh_data = pv.PolyData(vertices, facets)
                if part_mesh_info is None:
                    part_mesh_info = face_mesh_data
                else:
                    part_mesh_info = part_mesh_info.append_polydata(face_mesh_data)
        return part_mesh_info

    def _create_speos_feature_preview(
        self, plotter: Plotter, speos_feature: Union[SensorCamera, SensorRadiance, SensorIrradiance]
    ) -> Plotter:
        """Add speos feature visual preview to pyvista plotter object.

        Parameters
        ----------
        plotter: Plotter
            ansys.tools.visualization_interface.Plotter
        speos_feature: Union[SensorCamera, SensorRadiance, SensorIrradiance]
            speos feature whose visual data will be added.

        Returns
        -------
        Plotter
            ansys.tools.visualization_interface.Plotter
        """
        if not isinstance(speos_feature, (SensorIrradiance, SensorRadiance, SensorCamera)):
            return plotter
        plotter.plot(
            speos_feature.visual_data.data,
            show_edges=True,
            line_width=2,
            edge_color="red",
            color="orange",
            opacity=0.5,
        )
        match speos_feature:
            case SensorRadiance():
                plotter.plot(speos_feature.visual_data.coordinates.x_axis, color="red")
                plotter.plot(speos_feature.visual_data.coordinates.y_axis, color="green")
            case SensorIrradiance() | SensorCamera():
                plotter.plot(speos_feature.visual_data.coordinates.x_axis, color="red")
                plotter.plot(speos_feature.visual_data.coordinates.y_axis, color="green")
                plotter.plot(speos_feature.visual_data.coordinates.z_axis, color="blue")
        return plotter

    @graphics_required
    def _create_preview(self, viz_args=None) -> Plotter:
        """Create preview pyvista plotter object.

        Parameters
        ----------
        viz_args : dict
            contains arguments in dict format passed to add mesh function
            e.g.
            - {'style': 'wireframe'},
            - {'style': 'surface', 'color':'white'},
            - {'opacity': 0.7, 'color':'white', 'show_edges': False},
        """
        import pyvista as pv

        from ansys.tools.visualization_interface import Plotter

        if viz_args is None:
            viz_args = {}

        _preview_mesh = pv.PolyData()
        # Retrieve root part
        root_part_data = self.client[self.scene_link.get().part_guid].get()

        # Loop on all sub parts to retrieve their mesh
        if len(root_part_data.parts) != 0:
            for part_idx, part_item in enumerate(root_part_data.parts):
                part_item_data = self.client[part_item.part_guid].get()
                poly_data = self.__extract_part_mesh_info(
                    part_data=part_item_data,
                    part_coordinate_info=part_item.axis_system,
                )
                if poly_data is not None:
                    _preview_mesh = _preview_mesh.append_polydata(poly_data)

        # Add also the mesh of bodies directly contained in root part
        poly_data = self.__extract_part_mesh_info(part_data=root_part_data)
        if poly_data is not None:
            _preview_mesh = _preview_mesh.append_polydata(poly_data)
        p = Plotter()
        viz_args["show_edges"] = True
        p.plot(_preview_mesh, **viz_args)

        # Add speos visual data at the root part
        for feature in self._features:
            p = self._create_speos_feature_preview(p, feature)
        return p

    @graphics_required
    def preview(
        self,
        viz_args=None,
        screenshot: Optional[Union[str, Path]] = None,
    ) -> None:
        """Preview cad bodies inside the project's scene.

        Parameters
        ----------
        viz_args : dict
            contains arguments in dict format passed to add mesh function
            e.g.
            - {'style': 'wireframe'},
            - {'style': 'surface', 'color':'white'},
            - {'opacity': 0.7, 'color':'white', 'show_edges': False}.

        screenshot : str or Path or ``None``
            Path to save a screenshot of the plotter. If defined Plotter will only create the
            screenshot

        """
        if viz_args is None:
            viz_args = {"opacity": 1}
        if screenshot is not None:
            screenshot = Path(screenshot)

        p = self._create_preview(viz_args=viz_args)
        p.show(screenshot=screenshot)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module with method to process Protobuf messages."""

import json
from typing import Iterator, List

from google.protobuf.message import Message

from ansys.speos.core.kernel import SpeosClient, protobuf_message_to_dict


def dict_to_str(dict: dict) -> str:
    """Transform a dictionary into a string.

    Parameters
    ----------
    dict
        Dictionary to transform.

    Returns
    -------
    str
        String representation of the dictionary.
    """
    return json.dumps(dict, indent=4)


def _replace_guids(
    speos_client: SpeosClient, message: Message, ignore_simple_key: str = ""
) -> dict:
    # Transform protobuf message into dictionary
    json_dict = protobuf_message_to_dict(message=message)
    # Add for each element xxx_guid a key xxx with value the corresponding data from database
    _replace_guid_elt(
        speos_client=speos_client,
        json_dict=json_dict,
        ignore_simple_key=ignore_simple_key,
    )
    return json_dict


def _replace_guid_elt(
    speos_client: SpeosClient, json_dict: dict, ignore_simple_key: str = ""
) -> None:
    new_items = []
    for k, v in json_dict.items():
        # If we are in the case of key "xxx_guid", with a guid non empty
        # and that the key is not to ignore
        if k.endswith("_guid") and v != "" and k != ignore_simple_key:
            # Retrieve the item from db and transform it to dictionary
            new_v = protobuf_message_to_dict(message=speos_client[v].get())

            # This item can potentially have some "xxx_guid" fields to replace
            _replace_guid_elt(
                speos_client=speos_client,
                json_dict=new_v,
                ignore_simple_key=ignore_simple_key,
            )
            # Add the new value under "xxx" key
            new_items.append((k[: k.find("_guid")], new_v))
        # Possibility to have a list of guids : "xxx_guids"
        elif k.endswith("_guids") and len(v) != 0:
            # then a new list of values will be added under "xxxs" key
            new_key_list = k[: k.find("_guid")] + "s"
            new_value_list = []
            for iv in v:
                # Retrieve the item from db and transform it to dictionary
                new_v = protobuf_message_to_dict(message=speos_client[iv].get())

                # This item can potentially have some "xxx_guid" fields to replace
                _replace_guid_elt(
                    speos_client=speos_client,
                    json_dict=new_v,
                    ignore_simple_key=ignore_simple_key,
                )
                # Add the new value to the "xxxs" list
                new_value_list.append(new_v)
            new_items.append((new_key_list, new_value_list))

        # Call recursevely if the value is a dict or a list with dict as items values
        if isinstance(v, dict):
            _replace_guid_elt(
                speos_client=speos_client,
                json_dict=v,
                ignore_simple_key=ignore_simple_key,
            )
        elif isinstance(v, list):
            for iv in v:
                if isinstance(iv, dict):
                    _replace_guid_elt(
                        speos_client=speos_client,
                        json_dict=iv,
                        ignore_simple_key=ignore_simple_key,
                    )

    # To avoid modifying a dict when reading it, all changes were stored in new_items list
    # They are applied now
    for new_k, new_v in new_items:
        json_dict[new_k] = new_v


class _ReplacePropsElt:
    """Class to help replacing properties element."""

    def __init__(self) -> None:
        self.new_items = {}
        self.dict_to_complete = {}
        self.key_to_remove = ""
        self.dict_to_remove = {}


def _value_finder_key_startswith(dict_var: dict, key: str) -> Iterator[tuple[str, dict]]:
    """Find all (key,value) corresponding to a key that starts with input key."""
    for k, v in dict_var.items():
        if k.startswith(key):  # if the key fits the condition
            yield (k, v)

        if isinstance(v, dict):  # Look recursively if the value is a dict
            for kk, vv in _value_finder_key_startswith(dict_var=v, key=key):
                yield (kk, vv)


def _value_finder_key_endswith(dict_var: dict, key: str) -> Iterator[tuple[str, dict, dict]]:
    """Find all (key,value,parent dict) corresponding to a key that ends with input key."""
    for k, v in dict_var.items():
        if k.endswith(key):  # if the key fits the condition
            yield (k, v, dict_var)

        if isinstance(v, dict):  # Look recursively if the value is a dict
            for kk, vv, parent in _value_finder_key_endswith(dict_var=v, key=key):
                yield (kk, vv, parent)


def _replace_properties(json_dict: dict) -> None:
    """Replace every "xxx_properties" values to the place where the "xxx" value is defined."""
    replace_props_elts = []
    # Look for the "xxx_properties" elements
    for k, v, parent in _value_finder_key_endswith(dict_var=json_dict, key="_properties"):
        replace_props_elt = _ReplacePropsElt()
        replace_props_elt.key_to_remove = (
            k  # We will later remove this key, as its value were moved to correct place
        )
        replace_props_elt.dict_to_remove = (
            parent  # Remember the parent to be able to remove the key
        )

        # Look for the corresponding xxx item that we will complete with properties values
        for kk, vv in _value_finder_key_startswith(
            dict_var=json_dict, key=k[: k.find("_properties")]
        ):
            if kk != k and type(vv) is dict:
                replace_props_elt.dict_to_complete = (
                    vv  # Remember the dictionary to complete with properties
                )
                for kkk, vvv in v.items():
                    if not kkk.endswith("_properties"):
                        replace_props_elt.new_items[kkk] = vvv  # Store every property key, value
        replace_props_elts.append(replace_props_elt)

    # To avoid modifying dictionary while parsing it, modifications are done now
    for rpe in replace_props_elts:
        # Complete by adding key values
        for k, v in rpe.new_items.items():
            rpe.dict_to_complete[k] = v
        # Remove the key "xxx_properties" that is useless now
        if rpe.key_to_remove in rpe.dict_to_remove.keys():
            rpe.dict_to_remove.pop(rpe.key_to_remove)


def _finder_by_key(dict_var: dict, key: str, x_path: str = "") -> List[tuple[str, dict]]:
    """Find a key in a dictionary (recursively).

    It will return a list of (x_path, dictionary) of items corresponding to the key.
    """
    out_list = []

    # Loop on all dictionary items
    for k, v in dict_var.items():
        if k == key:  # If directly found, append to the out_list
            out_list.append(
                (x_path + "." + k, v)
            )  # x_path contains ".key" to be able to identify where to find the value returned
        elif isinstance(v, dict):
            # In case of a dictionary value, then call recursively _finder_by_key to check its keys
            x_path_bckp = x_path
            for x_path, vv in _finder_by_key(dict_var=v, key=key, x_path=x_path + "." + k):
                out_list.append((x_path, vv))
            x_path = x_path_bckp  # retrieve previous x_path

        elif isinstance(v, list):
            # In case of list, will loop on each list item to look for whished key
            x_path_bckp = x_path
            x_path += (
                "." + k + "["
            )  # x_path contains .key[idx] to be able to identify where to find the value returned
            i = 0
            for item in v:
                if isinstance(item, dict):
                    x_path_bckp2 = x_path
                    # In case the dict has field name, use it in x_path like .key[.name='TheName']
                    # it is more meaningful that just [idx]
                    if "name" in item.keys():
                        x_path = x_path + ".name='" + item["name"] + "']"
                    else:  # if no field name, then just use .key[idx]
                        x_path = x_path + str(i) + "]"

                    # Call recursively _finder_by_key to check items keys
                    for x_path, vv in _finder_by_key(dict_var=item, key=key, x_path=x_path):
                        out_list.append((x_path, vv))
                    x_path = x_path_bckp2  # retrieve previous x_path
                i = i + 1
            x_path = x_path_bckp  # retrieve previous x_path
    return out_list


def _flatten_dict(dict_var: dict):
    """Flatten a dictionary (recursively).

    It will return a dictionary of (key, value) of items corresponding to the key.
    If a same key appears multiple time, the last value will be used as the value.

    Parameters
    ----------
    dict_var: dict

    Returns
    -------
    dict

    """
    flat_dict = {}
    for k, v in dict_var.items():
        if isinstance(v, dict):
            flat_dict.update(_flatten_dict(v))
        flat_dict[k] = v
    return flat_dict



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a way to interact with Speos BSDF file."""

from __future__ import annotations

from collections import Counter, UserDict
from collections.abc import Collection
from pathlib import Path
from typing import Union
import warnings

from google.protobuf.empty_pb2 import Empty
import grpc
import numpy as np

import ansys.api.speos.bsdf.v1.anisotropic_bsdf_pb2 as anisotropic_bsdf__v1__pb2
import ansys.api.speos.bsdf.v1.anisotropic_bsdf_pb2_grpc as anisotropic_bsdf__v1__pb2_grpc
import ansys.api.speos.bsdf.v1.bsdf_creation_pb2 as bsdf_creation__v1__pb2
import ansys.api.speos.bsdf.v1.bsdf_creation_pb2_grpc as bsdf_creation__v1__pb2_grpc
import ansys.api.speos.bsdf.v1.spectral_bsdf_pb2 as spectral_bsdf__v1__pb2
import ansys.api.speos.bsdf.v1.spectral_bsdf_pb2_grpc as spectral_bsdf__v1__pb2_grpc
import ansys.speos.core
from ansys.speos.core.speos import Speos


class BaseBSDF:
    """Super class for all BSDF datamodels.

    Parameters
    ----------
    speos : ansys.speos.core.speos.Speos
    stub :
        grpc stub to connect to BSDF service
    namespace :
        grpc namespace for the bsdf

    Notes
    -----
    This is a Super class, **Do not instantiate this class yourself**

    """

    def __init__(self, speos: Speos, stub, namespace):
        self.client = speos.client
        self._stub = stub
        self._grpcbsdf = None
        self.has_transmission = False
        self.has_reflection = False
        self.anisotropy_vector = [1, 0, 0]
        self.description = ""
        self.__namespace = namespace
        self.__interpolation_settings = None

    @property
    def has_transmission(self) -> bool:
        """Contains the BSDF Transmission data."""
        return self._has_transmission

    @has_transmission.setter
    def has_transmission(self, value: bool):
        if value:
            self._has_transmission = value
        else:
            self._has_transmission = value
            self._btdf = None

    @property
    def has_reflection(self) -> bool:
        """Contains the BSDF Reflection data."""
        return self._has_reflection

    @has_reflection.setter
    def has_reflection(self, value: bool):
        if value:
            self._has_reflection = value
        else:
            self._has_reflection = value
            self._brdf = None

    @property
    def brdf(self) -> Collection[BxdfDatapoint]:
        """List of BRDFDatapoints."""
        return self._brdf

    @brdf.setter
    def brdf(self, value: list[BxdfDatapoint]):
        if value is None:
            self._brdf = value
        else:
            check = any([bxdf.is_brdf for bxdf in value])
            if check is True:
                value.sort(key=lambda x: (x.anisotropy, x.wavelength, x.incident_angle))
                self._brdf = value
                self.has_reflection = True
            else:
                raise ValueError("One or multiple datapoints are transmission datapoints")

    @property
    def btdf(self) -> Collection[BxdfDatapoint]:
        """List of BTDFDatapoints."""
        return self._btdf

    @btdf.setter
    def btdf(self, value: list[BxdfDatapoint]):
        if value is None:
            self._btdf = value
        else:
            check = any([not bxdf.is_brdf for bxdf in value])
            if check is True:
                value.sort(key=lambda x: (x.anisotropy, x.wavelength, x.incident_angle))
                self._btdf = value
                self.has_transmission = True
            else:
                raise ValueError("One or multiple datapoints are reflection datapoints")

    @property
    def nb_incidents(self) -> list[int]:
        """Number of incidence angle for reflection and transmission.

        Returns
        -------
        list[int]:
            first value of the list is nb of reflective data, second value is nb of transmittive
            data
        """
        if self.has_transmission:
            t = len(self.btdf)
        else:
            t = 0
        if self.has_reflection:
            r = len(self.brdf)
        else:
            r = 0
        return [r, t]

    @property
    def incident_angles(self) -> list[Union[list[float], None], Union[list[float], None]]:
        """List of incident angles for reflection and transmission.

        Returns
        -------
        list[Union[list[float], None],Union[list[float], None]]
            Returns a nested list of incidence angels for reflective and transmittive
            data if not available the value will be None
        """
        if self.has_transmission:
            t_angle = [bxdf.incident_angle for bxdf in self.btdf]
        else:
            t_angle = []
        if self.has_reflection:
            r_angle = [bxdf.incident_angle for bxdf in self.brdf]
        else:
            r_angle = []
        return [r_angle, t_angle]

    @property
    def interpolation_settings(self) -> Union[None, InterpolationEnhancement]:
        """Interpolation enhancement settings of the bsdf file.

        If bsdf file does not have interpolation enhancement settings, return None.
        if bsdf file has interpolation enhancement settings, return InterpolationEnhancement.
        """
        return self.__interpolation_settings

    def create_interpolation_enhancement(
        self, index_1: float = 1.0, index_2: float = 1.0
    ) -> InterpolationEnhancement:
        """Apply automatic interpolation enhancement.

        Return interpolation settings to user if settings need change.

        Parameters
        ----------
        index_1 : float
            outside refractive index
        index_2 : float
            inside refractive index

        Returns
        -------
        ansys.speos.core.bsdf._InterpolationEnhancement
            automatic interpolation settings with index_1 = 1 and index_2 = 1 by default.
        """
        self._stub.Import(self._grpcbsdf)
        self.__interpolation_settings = InterpolationEnhancement(
            bsdf=self, bsdf_namespace=self.__namespace, index_1=index_1, index_2=index_2
        )
        return self.__interpolation_settings


class InterpolationEnhancement:
    """Class to facilitate Specular interpolation enhancement.

    Notes
    -----
    **Do not instantiate this class yourself**
    """

    class _InterpolationSettings(UserDict):
        """Class to facilitate interpolation settings as fixed dictionary."""

        def __init__(self, initial_data) -> None:
            self._fixed_keys = list(initial_data.keys())
            super().__init__(initial_data)

        def __setitem__(self, key, value) -> None:
            """Dis-able setting values under two conditions.

            1) methods to set value if dictionary value is a fixed dictionary.
            2) key is not inside fixed dictionary.
            """
            if key in self._fixed_keys:
                if not isinstance(self.get(key), self.__class__):
                    super().__setitem__(key, value)
                else:
                    raise ValueError(
                        "Cannot update key {key} with a FixedKeyDict as value.".format(key=key)
                    )
            else:
                raise KeyError("Cannot add new key: {key} is not allowed.".format(key=key))

        def __delitem__(self, key) -> None:
            """Dis-able delete key from a fixed dictionary."""
            raise KeyError("Deletion of key: {key}  is not allowed.".format(key=key))

        def __iter__(self):
            """Iterate keys in fixed key order."""
            return iter(self._fixed_keys)

        def keys(self) -> list[str]:
            """Return keys from fixed dictionary."""
            return self._fixed_keys

        def items(self):
            """Return items from fixed dictionary."""
            return [(key, self[key]) for key in self._fixed_keys]

        def update(self, *args, **kwargs) -> None:
            """
            Update the fixed dictionary with multiple values.

            Example:
                my_dict.update({'a': 10, 'b': 20})
                my_dict.update({'x': 100})
                my_dict.update(x=200, y=300)

            Parameters
            ----------
            args: tuple
                Dictionary or iterable of key-value pairs.
            kwargs: dict
                Key-value pairs to update the dictionary.
            """
            updates = dict(*args, **kwargs)
            for key in updates:
                if key not in self._fixed_keys:
                    raise KeyError("Cannot add new key: {key} is not allowed.".format(key=key))
                if isinstance(self.get(key), self.__class__):
                    raise ValueError(
                        "Cannot update key {key} with FixedKeyDict as value.".format(key=key)
                    )
            super().update(updates)

    def __init__(
        self,
        bsdf: Union[AnisotropicBSDF, SpectralBRDF],
        bsdf_namespace: Union[spectral_bsdf__v1__pb2, anisotropic_bsdf__v1__pb2],
        index_1: Union[float, None] = 1.0,
        index_2: Union[float, None] = 1.0,
    ) -> None:
        self._bsdf = bsdf

        self.__cones_data = None
        try:
            self.__cones_data = self._bsdf._stub.GetSpecularInterpolationEnhancementData(Empty())
            if index_1 is not None and index_2 is not None:
                self.__indices = bsdf_namespace.RefractiveIndices(
                    refractive_index_1=index_1,
                    refractive_index_2=index_2,
                )
                self._bsdf._stub.GenerateSpecularInterpolationEnhancementData(self.__indices)
                self.__cones_data = self._bsdf._stub.GetSpecularInterpolationEnhancementData(
                    Empty()
                )
        except grpc.RpcError:
            self.__indices = bsdf_namespace.RefractiveIndices(
                refractive_index_1=index_1,
                refractive_index_2=index_2,
            )
            self._bsdf._stub.GenerateSpecularInterpolationEnhancementData(self.__indices)
            self.__cones_data = self._bsdf._stub.GetSpecularInterpolationEnhancementData(Empty())

    @property
    def index1(self) -> float:
        """Refractive index on reflection side."""
        return self.__cones_data.refractive_index_1

    @index1.setter
    def index1(self, value: Union[float, int]) -> None:
        """Set refractive index on reflection side."""
        self.__cones_data.refractive_index_1 = value
        self._bsdf._stub.Import(self._bsdf._grpcbsdf)
        self._bsdf._stub.SetSpecularInterpolationEnhancementData(self.__cones_data)

    @property
    def index2(self) -> float:
        """Refractive index on transmission side."""
        return self.__cones_data.refractive_index_2

    @index2.setter
    def index2(self, value: Union[float, int]) -> None:
        """Set refractive index on transmission side."""
        self.__cones_data.refractive_index_2 = value
        self._bsdf._stub.Import(self._bsdf._grpcbsdf)
        self._bsdf._stub.SetSpecularInterpolationEnhancementData(self.__cones_data)

    @property
    def get_reflection_interpolation_settings(self) -> Union[None, _InterpolationSettings]:
        """Return a fixed dictionary for reflection interpolation settings to be set by user."""
        if not self._bsdf.has_reflection:
            return None
        if isinstance(self._bsdf, AnisotropicBSDF):
            reflection_interpolation_settings = self._InterpolationSettings(
                {str(key): 0 for key in self._bsdf.anisotropic_angles[0]}
            )
            for aniso_sample_index, ani_sample in enumerate(
                self.__cones_data.reflection.anisotropic_samples
            ):
                reflection_incident_interpolation_settings = self._InterpolationSettings(
                    {str(key): 0 for key in self._bsdf.incident_angles[0]}
                )
                tmp_reflection_key = str(self._bsdf.anisotropic_angles[0][aniso_sample_index])
                reflection_interpolation_settings.update(
                    {tmp_reflection_key: reflection_incident_interpolation_settings}
                )
                for incident_sample_index, incident in enumerate(ani_sample.incidence_samples):
                    tmp_reflection_incident_key = str(
                        self._bsdf.incident_angles[0][
                            aniso_sample_index * len(ani_sample.incidence_samples)
                            + incident_sample_index
                        ]
                    )
                    reflection_interpolation_settings[
                        str(self._bsdf.anisotropic_angles[0][aniso_sample_index])
                    ].update(
                        {
                            tmp_reflection_incident_key: {
                                "half_angle": incident.cone_half_angle,
                                "height": incident.cone_height,
                            }
                        }
                    )
            return reflection_interpolation_settings
        if isinstance(self._bsdf, SpectralBRDF):
            reflection_interpolation_settings = self._InterpolationSettings(
                {str(key): 0 for key in self._bsdf.wavelength}
            )
            r_angles = list(set(self._bsdf.incident_angles[0]))
            for wl_index, wl in enumerate(self._bsdf.wavelength):
                tmp_reflection_key = str(wl)
                reflection_incident_interpolation_settings = self._InterpolationSettings(
                    {str(key): 0 for key in r_angles}
                )
                reflection_interpolation_settings.update(
                    {tmp_reflection_key: reflection_incident_interpolation_settings}
                )
                for inc_index, inc in enumerate(r_angles):
                    ani_sample = self.__cones_data.wavelength_incidence_samples[
                        (wl_index + 1) * inc_index
                    ]
                    tmp_reflection_incident_key = str(inc)
                    reflection_interpolation_settings[str(wl)].update(
                        {
                            tmp_reflection_incident_key: {
                                "half_angle": ani_sample.reflection.cone_half_angle,
                                "height": ani_sample.reflection.cone_height,
                            }
                        }
                    )
            return reflection_interpolation_settings
        else:
            raise ValueError("only anisotropic and spectral bsdf supported")

    def set_interpolation_settings(
        self, is_brdf: bool, settings: InterpolationEnhancement._InterpolationSettings
    ) -> None:
        """Set interpolation obtained from bsdf Class or that modified by user.

        Parameters
        ----------
        is_brdf: bool
            true if settings is for brdf, else for btdf
        settings: InterpolationEnhancement._InterpolationSettings
            interpolation settings.
        """
        if not isinstance(
            settings, ansys.speos.core.bsdf.InterpolationEnhancement._InterpolationSettings
        ):
            raise ImportError("only interpolation settings are supported")
        if is_brdf and not self._bsdf.has_reflection:
            raise ValueError("BSDF has no reflection data")
        if not is_brdf and not self._bsdf.has_transmission:
            raise ValueError("BSDF has no transmission data")
        if isinstance(self._bsdf, AnisotropicBSDF):
            self._bsdf._stub.Import(self._bsdf._grpcbsdf)
            if is_brdf:
                for iso_sample_key_index, iso_sample_key in enumerate(settings.keys()):
                    for incident_key_index, incident_key in enumerate(
                        settings[iso_sample_key].keys()
                    ):
                        self.__cones_data.reflection.anisotropic_samples[
                            iso_sample_key_index
                        ].incidence_samples[incident_key_index].cone_half_angle = settings[
                            iso_sample_key
                        ][incident_key]["half_angle"]
                        self.__cones_data.reflection.anisotropic_samples[
                            iso_sample_key_index
                        ].incidence_samples[incident_key_index].cone_height = settings[
                            iso_sample_key
                        ][incident_key]["height"]
                self._bsdf._stub.SetSpecularInterpolationEnhancementData(self.__cones_data)
            else:
                for iso_sample_key_index, iso_sample_key in enumerate(settings.keys()):
                    for incident_key_index, incident_key in enumerate(
                        settings[iso_sample_key].keys()
                    ):
                        self.__cones_data.transmission.anisotropic_samples[
                            iso_sample_key_index
                        ].incidence_samples[incident_key_index].cone_half_angle = settings[
                            iso_sample_key
                        ][incident_key]["half_angle"]
                        self.__cones_data.transmission.anisotropic_samples[
                            iso_sample_key_index
                        ].incidence_samples[incident_key_index].cone_height = settings[
                            iso_sample_key
                        ][incident_key]["height"]
                self._bsdf._stub.SetSpecularInterpolationEnhancementData(self.__cones_data)
        elif isinstance(self._bsdf, SpectralBRDF):
            self._bsdf._stub.Import(self._bsdf._grpcbsdf)
            if is_brdf:
                for wl_sample_key_index, wl_sample_key in enumerate(settings.keys()):
                    for incident_key_index, incident_key in enumerate(
                        settings[wl_sample_key].keys()
                    ):
                        self.__cones_data.wavelength_incidence_samples[
                            (wl_sample_key_index + 1) * incident_key_index
                        ].reflection.cone_half_angle = settings[wl_sample_key][incident_key][
                            "half_angle"
                        ]
                        self.__cones_data.wavelength_incidence_samples[
                            (wl_sample_key_index + 1) * incident_key_index
                        ].reflection.cone_height = settings[wl_sample_key][incident_key]["height"]
                self._bsdf._stub.SetSpecularInterpolationEnhancementData(self.__cones_data)
            else:
                for wl_sample_key_index, wl_sample_key in enumerate(settings.keys()):
                    for incident_key_index, incident_key in enumerate(
                        settings[wl_sample_key].keys()
                    ):
                        self.__cones_data.wavelength_incidence_samples[
                            (wl_sample_key_index + 1) * incident_key_index
                        ].transmission.cone_half_angle = settings[wl_sample_key][incident_key][
                            "half_angle"
                        ]
                        self.__cones_data.wavelength_incidence_samples[
                            (wl_sample_key_index + 1) * incident_key_index
                        ].transmission.cone_height = settings[wl_sample_key][incident_key]["height"]
                self._bsdf._stub.SetSpecularInterpolationEnhancementData(self.__cones_data)
        else:
            raise ValueError("only anisotropic bsdf and spectral brdf are supported")

    @property
    def get_transmission_interpolation_settings(self) -> Union[None, _InterpolationSettings]:
        """Return a fixed dictionary for reflection interpolation settings to be set by user."""
        if not self._bsdf.has_transmission:
            return None
        if isinstance(self._bsdf, AnisotropicBSDF):
            transmission_interpolation_settings = self._InterpolationSettings(
                {str(key): 0 for key in self._bsdf.anisotropic_angles[1]}
            )
            for aniso_sample_index, ani_sample in enumerate(
                self.__cones_data.transmission.anisotropic_samples
            ):
                transmission_incident_interpolation_settings = self._InterpolationSettings(
                    {str(key): 0 for key in self._bsdf.incident_angles[1]}
                )
                tmp_transmission_key = str(self._bsdf.anisotropic_angles[1][aniso_sample_index])
                transmission_interpolation_settings.update(
                    {tmp_transmission_key: transmission_incident_interpolation_settings}
                )
                for incident_sample_index, incident in enumerate(ani_sample.incidence_samples):
                    tmp_transmission_incident_key = str(
                        self._bsdf.incident_angles[1][
                            aniso_sample_index * len(ani_sample.incidence_samples)
                            + incident_sample_index
                        ]
                    )
                    transmission_interpolation_settings[
                        str(self._bsdf.anisotropic_angles[1][aniso_sample_index])
                    ].update(
                        {
                            tmp_transmission_incident_key: {
                                "half_angle": incident.cone_half_angle,
                                "height": incident.cone_height,
                            }
                        }
                    )
            return transmission_interpolation_settings
        if isinstance(self._bsdf, SpectralBRDF):
            transmission_interpolation_settings = self._InterpolationSettings(
                {str(key): 0 for key in self._bsdf.wavelength}
            )
            r_angles = list(set(self._bsdf.incident_angles[0]))
            for wl_index, wl in enumerate(self._bsdf.wavelength):
                tmp_reflection_key = str(wl)
                transmission_incident_interpolation_settings = self._InterpolationSettings(
                    {str(key): 0 for key in r_angles}
                )
                transmission_interpolation_settings.update(
                    {tmp_reflection_key: transmission_incident_interpolation_settings}
                )
                for inc_index, inc in enumerate(r_angles):
                    sample = self.__cones_data.wavelength_incidence_samples[
                        (wl_index + 1) * inc_index
                    ]
                    tmp_reflection_incident_key = str(inc)
                    transmission_interpolation_settings[str(wl)].update(
                        {
                            tmp_reflection_incident_key: {
                                "half_angle": sample.transmission.cone_half_angle,
                                "height": sample.transmission.cone_height,
                            }
                        }
                    )
            return transmission_interpolation_settings
        else:
            raise ValueError("only anisotropic and spectral bsdf supported")


class AnisotropicBSDF(BaseBSDF):
    """BSDF - Bidirectional scattering distribution function.

    This class contains the methods and functions to load and edit existing Speos bsdf datasets.

    Parameters
    ----------
    speos : ansys.speos.core.speos.Speos
        Speos Object to connect to speos rpc server
    file_path : Union[Path, str]
        File path to bsdf file
    """

    def __init__(self, speos: Speos, file_path: Union[Path, str] = None):
        super().__init__(
            speos,
            anisotropic_bsdf__v1__pb2_grpc.AnisotropicBsdfServiceStub(speos.client.channel),
            anisotropic_bsdf__v1__pb2,
        )
        self._spectrum_incidence = [0, 0]
        self._spectrum_anisotropy = [0, 0]
        if file_path:
            file_path = Path(file_path)
            self._grpcbsdf = self._import_file(file_path)
            self._brdf, self._btdf = self._extract_bsdf()
            self._has_transmission = bool(self._btdf)
            self._has_reflection = bool(self._brdf)
            self._reflection_spectrum, self._transmission_spectrum = self._extract_spectrum()
            try:
                self._stub.GetSpecularInterpolationEnhancementData(Empty())
                self._BaseBSDF__interpolation_settings = InterpolationEnhancement(
                    bsdf=self,
                    bsdf_namespace=anisotropic_bsdf__v1__pb2,
                    index_1=None,
                    index_2=None,
                )
            except grpc.RpcError:
                self.__interpolation_settings = None
        else:
            self._transmission_spectrum, self._reflection_spectrum = None, None

    def get(self, key=""):
        """Retrieve any information from the BSDF object.

        Parameters
        ----------
        key : str
            Name of the property.

        Returns
        -------
        property
            Values/content of the associated property.
        """
        data = {k: v.fget(self) for k, v in BaseBSDF.__dict__.items() if isinstance(v, property)}
        data.update(
            {
                k: v.fget(self)
                for k, v in AnisotropicBSDF.__dict__.items()
                if isinstance(v, property)
            }
        )
        if key == "":
            return data
        elif data.get(key):
            return data.get(key)
        else:
            print("Used key: {} not found in key list: {}.".format(key, data.keys()))

    def __str__(self):
        """Create string representation of a BSDF."""
        return str(self.get())

    def _import_file(self, filepath):
        file_name = anisotropic_bsdf__v1__pb2.FileName()
        file_name.file_name = str(filepath)
        self._stub.Load(file_name)
        return self._stub.Export(Empty())

    def _extract_bsdf(self) -> tuple[Collection[BxdfDatapoint], Collection[BxdfDatapoint]]:
        self.description = self._grpcbsdf.description
        self.ansistropy_vector = [
            self._grpcbsdf.anisotropy_vector.x,
            self._grpcbsdf.anisotropy_vector.y,
            self._grpcbsdf.anisotropy_vector.z,
        ]
        brdf = []
        btdf = []
        for ani_bsdf_data in self._grpcbsdf.reflection.anisotropic_samples:
            anisotropic_angle = ani_bsdf_data.anisotropic_sample
            for bsdf_data in ani_bsdf_data.incidence_samples:
                incident_angle = bsdf_data.incidence_sample
                thetas = np.array(bsdf_data.theta_samples)
                phis = np.array(bsdf_data.phi_samples)
                bsdf = np.array(bsdf_data.bsdf_cos_theta).reshape((len(thetas), len(phis)))
                tis = bsdf_data.integral
                brdf.append(
                    BxdfDatapoint(True, incident_angle, thetas, phis, bsdf, tis, anisotropic_angle)
                )
        for ani_bsdf_data in self._grpcbsdf.transmission.anisotropic_samples:
            anisotropic_angle = ani_bsdf_data.anisotropic_sample
            for bsdf_data in ani_bsdf_data.incidence_samples:
                incident_angle = bsdf_data.incidence_sample
                thetas = np.array(bsdf_data.theta_samples)
                phis = np.array(bsdf_data.phi_samples)
                bsdf = np.array(bsdf_data.bsdf_cos_theta).reshape((len(thetas), len(phis)))
                tis = bsdf_data.integral
                btdf.append(
                    BxdfDatapoint(False, incident_angle, thetas, phis, bsdf, tis, anisotropic_angle)
                )
        return brdf, btdf

    def _extract_spectrum(self) -> list[np.ndarray, np.ndarray]:
        if self.has_reflection:
            self.spectrum_incidence[0] = self._grpcbsdf.reflection.spectrum_incidence
            self.spectrum_anisotropy[0] = self._grpcbsdf.reflection.spectrum_anisotropy
        if self.has_transmission:
            self.spectrum_incidence[1] = self._grpcbsdf.transmission.spectrum_incidence
            self.spectrum_anisotropy[1] = self._grpcbsdf.transmission.spectrum_anisotropy
        refl_s = np.array([[], []])
        trans_s = np.array([[], []])
        for value in self._grpcbsdf.reflection.spectrum:
            refl_s = np.append(refl_s, [[value.wavelength], [value.coefficient]], axis=1)
        for value in self._grpcbsdf.transmission.spectrum:
            trans_s = np.append(trans_s, [[value.wavelength], [value.coefficient]], axis=1)
        return [refl_s, trans_s]

    @property
    def anisotropic_angles(self):
        """Anisotropic angles available in bsdf data."""
        if self.has_transmission:
            t_angles = [btdf.anisotropy for btdf in self.btdf]
        else:
            t_angles = []
        if self.has_reflection:
            r_angles = [brdf.anisotropy for brdf in self.brdf]
        else:
            r_angles = []
        return [list(Counter(r_angles).keys()), list(Counter(t_angles).keys())]

    @property
    def spectrum_incidence(self) -> list[float]:
        """Incident angle (theta) of spectrum measurement.

        First value is for reflection second for transmission
        """
        return self._spectrum_incidence

    @spectrum_incidence.setter
    def spectrum_incidence(self, value) -> list[float]:
        if isinstance(value, float) and self.has_reflection and self.has_transmission:
            raise ValueError("You need to define the value for both reflection and transmission")
        elif isinstance(value, float) and 0 <= value <= np.pi / 2:
            if self.has_reflection:
                self._spectrum_incidence[0] = value
            else:
                self._spectrum_incidence[1] = value
        elif isinstance(value, list):
            if len(value) == 2 and any([0 <= theta <= np.pi / 2 for theta in value]):
                self._spectrum_incidence = value
        else:
            raise ValueError(
                "You need to define the value in radian for both reflection and transmission"
            )

    @property
    def spectrum_anisotropy(self) -> list[float]:
        """Incident angle (phi) of spectrum measurement.

        First value is for reflection second for transmission
        """
        return self._spectrum_anisotropy

    @spectrum_anisotropy.setter
    def spectrum_anisotropy(self, value):
        if isinstance(value, float) and self.has_reflection and self.has_transmission:
            raise ValueError("You need to define the value for both reflection and transmission")
        elif isinstance(value, float) and 0 <= value <= 2 * np.pi:
            if self.has_reflection:
                self._spectrum_anisotropy[0] = value
            else:
                self._spectrum_anisotropy[1] = value
        elif isinstance(value, list):
            if len(value) == 2 and any([0 <= theta <= 2 * np.pi for theta in value]):
                self._spectrum_anisotropy = value
        else:
            raise ValueError(
                "You need to define the value in radian for both reflection and transmission"
            )

    @property
    def reflection_spectrum(self):
        """Reflection Spectrum of the bsdf.

        The spectrum is used to modulate the bsdf.
        """
        return self._reflection_spectrum

    @reflection_spectrum.setter
    def reflection_spectrum(self, value: list[Collection[float], Collection[float]]):
        if len(value[0]) == len(value[1]):
            self._reflection_spectrum = value
        else:
            raise ValueError("You need the same number of wavelength and energy values")

    @property
    def transmission_spectrum(self):
        """Transmission  Spectrum of the bsdf.

        The spectrum is used to modulate the bsdf.
        """
        return self._transmission_spectrum

    @transmission_spectrum.setter
    def transmission_spectrum(self, value: list[Collection[float], Collection[float]]):
        if len(value[0]) == len(value[1]):
            self._transmission_spectrum = value
        else:
            raise ValueError("You need the same number of wavelength and energy values")

    def reset(self):
        """Reset BSDF data to what was stored in file."""
        self._brdf, self._btdf = self._extract_bsdf()
        self._has_transmission = bool(self._btdf)
        self._has_reflection = bool(self._brdf)
        self._transmission_spectrum, self._reflection_spectrum = self._extract_spectrum()

    def commit(self):
        """Sent Data to gRPC interface."""
        # set basic values
        bsdf = anisotropic_bsdf__v1__pb2.AnisotropicBsdfData()
        bsdf.description = self.description
        bsdf.anisotropy_vector.x = self.anisotropy_vector[0]
        bsdf.anisotropy_vector.y = self.anisotropy_vector[1]
        bsdf.anisotropy_vector.z = self.anisotropy_vector[2]
        if self.has_reflection:
            bsdf.reflection.spectrum_incidence = self.spectrum_incidence[0]
            bsdf.reflection.spectrum_anisotropy = self.spectrum_anisotropy[0]
            for w in range(len(self.reflection_spectrum[0])):
                pair = bsdf.reflection.spectrum.add()
                pair.wavelength = self.reflection_spectrum[0][w]
                pair.coefficient = self.reflection_spectrum[1][w]
            for ani in self.anisotropic_angles[0]:
                slice = bsdf.reflection.anisotropic_samples.add()
                slice.anisotropic_sample = ani
                for brdf in self.brdf:
                    if brdf.anisotropy == ani:
                        incidence_diag = slice.incidence_samples.add()
                        incidence_diag.incidence_sample = brdf.incident_angle
                        # intensity diagrams
                        incidence_diag.phi_samples[:] = list(brdf.phi_values)
                        incidence_diag.theta_samples[:] = list(brdf.theta_values)
                        incidence_diag.bsdf_cos_theta[:] = brdf.bxdf.flatten().tolist()
        if self.has_transmission:
            bsdf.transmission.spectrum_incidence = self.spectrum_incidence[1]
            bsdf.transmission.spectrum_anisotropy = self.spectrum_anisotropy[1]
            for w in range(len(self.transmission_spectrum[0])):
                pair = bsdf.transmission.spectrum.add()
                pair.wavelength = self.transmission_spectrum[0][w]
                pair.coefficient = self.transmission_spectrum[1][w]
            for ani in self.anisotropic_angles[1]:
                slice = bsdf.transmission.anisotropic_samples.add()
                slice.anisotropic_sample = ani
                for btdf in self.btdf:
                    if btdf.anisotropy == ani:
                        incidence_diag = slice.incidence_samples.add()
                        incidence_diag.incidence_sample = btdf.incident_angle
                        # intensity diagrams
                        incidence_diag.phi_samples[:] = list(btdf.phi_values)
                        incidence_diag.theta_samples[:] = list(btdf.theta_values)
                        incidence_diag.bsdf_cos_theta[:] = btdf.bxdf.flatten().tolist()
        self._stub.Import(bsdf)
        self._grpcbsdf = bsdf
        if self._BaseBSDF__interpolation_settings is not None:
            self._stub.SetSpecularInterpolationEnhancementData(
                self._BaseBSDF__interpolation_settings._InterpolationEnhancement__cones_data
            )

    def save(self, file_path: Union[Path, str], commit: bool = True) -> Path:
        """Save a Speos anistropic bsdf.

        Parameters
        ----------
        file_path : Union[Path, str]
            Filepath to save bsdf
        commit : bool
            commit data before saving

        Returns
        -------
        Path
            File location
        """
        file_path = Path(file_path)
        file_name = anisotropic_bsdf__v1__pb2.FileName()
        if commit:
            self.commit()
        else:
            self._stub.Import(self._grpcbsdf)
        if file_path.suffix == ".anisotropicbsdf":
            file_name.file_name = str(file_path)
        else:
            file_name.file_name = str(file_path.parent / (file_path.name + ".anisotropicbsdf"))
        self._stub.Save(file_name)
        return Path(file_name.file_name)


class SpectralBRDF(BaseBSDF):
    """BSDF - Bidirectional scattering distribution function.

    This class contains the methods and functions to load and edit existing Speos bsdf datasets.

    Parameters
    ----------
    speos : ansys.speos.core.speos.Speos
        Speos Object to connect to speos rpc server
    file_path : Union[Path, str]
        File path to bsdf file
    """

    def __init__(self, speos: Speos, file_path: Union[Path, str] = None):
        super().__init__(
            speos,
            spectral_bsdf__v1__pb2_grpc.SpectralBsdfServiceStub(speos.client.channel),
            spectral_bsdf__v1__pb2,
        )
        self._spectrum_incidence = [0, 0]
        self._spectrum_anisotropy = [0, 0]
        if file_path:
            file_path = Path(file_path)
            self._grpcbsdf = self._import_file(file_path)
            self.brdf, self.btdf = self._extract_bsdf()
            self._has_transmission = bool(self._btdf)
            self._has_reflection = bool(self._brdf)
            try:
                self._stub.GetSpecularInterpolationEnhancementData(Empty())
                self._BaseBSDF__interpolation_settings = InterpolationEnhancement(
                    bsdf=self,
                    bsdf_namespace=spectral_bsdf__v1__pb2,
                    index_1=None,
                    index_2=None,
                )
            except grpc.RpcError:
                self.__interpolation_settings = None
        else:
            self._transmission_spectrum, self._reflection_spectrum = None, None

    @property
    def wavelength(self):
        """List of all Wavelength in BRDF."""
        r_wl = []
        t_wl = []
        if self.has_reflection:
            for brdf in self.brdf:
                r_wl.append(brdf.wavelength)
            return list(set(r_wl))
        if self.has_transmission:
            for btdf in self.btdf:
                t_wl.append(btdf.wavelength)
            return list(set(t_wl))
        else:
            return []

    def get(self, key=""):
        """Retrieve any information from the BSDF object.

        Parameters
        ----------
        key : str
            Name of the property.

        Returns
        -------
        property
            Values/content of the associated property.
        """
        data = {k: v.fget(self) for k, v in BaseBSDF.__dict__.items() if isinstance(v, property)}
        data.update(
            {k: v.fget(self) for k, v in SpectralBRDF.__dict__.items() if isinstance(v, property)}
        )
        if key == "":
            return data
        elif data.get(key):
            return data.get(key)
        else:
            print("Used key: {} not found in key list: {}.".format(key, data.keys()))

    def __str__(self):
        """Create string representation of a BSDF."""
        return str(self.get())

    def _import_file(self, filepath):
        file_name = spectral_bsdf__v1__pb2.FileName()
        file_name.file_name = str(filepath)
        self._stub.Load(file_name)
        return self._stub.Export(Empty())

    def _extract_bsdf(self) -> tuple[list[BxdfDatapoint], list[BxdfDatapoint]]:
        self.description = self._grpcbsdf.description
        brdf = []
        btdf = []
        for i, spectral_bsdf_data in enumerate(self._grpcbsdf.wavelength_incidence_samples):
            anisotropic_angle = 0
            incident_angle = self._grpcbsdf.incidence_samples[
                i % len(self._grpcbsdf.incidence_samples)
            ]
            wl = self._grpcbsdf.wavelength_samples[int(i / len(self._grpcbsdf.incidence_samples))]
            if spectral_bsdf_data.HasField("reflection"):
                thetas = np.array(spectral_bsdf_data.reflection.theta_samples)
                phis = np.array(spectral_bsdf_data.reflection.phi_samples)
                bsdf = np.array(spectral_bsdf_data.reflection.bsdf_cos_theta).reshape(
                    (len(thetas), len(phis))
                )
                tis = spectral_bsdf_data.reflection.integral
                brdf.append(
                    BxdfDatapoint(
                        True, incident_angle, thetas, phis, bsdf, tis, anisotropic_angle, wl
                    )
                )
            if spectral_bsdf_data.HasField("transmission"):
                thetas = np.array(spectral_bsdf_data.transmission.theta_samples)
                phis = np.array(spectral_bsdf_data.transmission.phi_samples)
                bsdf = np.array(spectral_bsdf_data.transmission.bsdf_cos_theta).reshape(
                    (len(thetas), len(phis))
                )
                tis = spectral_bsdf_data.transmission.integral
                btdf.append(
                    BxdfDatapoint(
                        False, incident_angle, thetas, phis, bsdf, tis, anisotropic_angle, wl
                    )
                )
        if not brdf:
            brdf = None
        if not btdf:
            btdf = None
        return brdf, btdf

    def reset(self):
        """Reset BSDF data to what was stored in file."""
        self._brdf, self._btdf = self._extract_bsdf()
        self._has_transmission = bool(self._btdf)
        self._has_reflection = bool(self._brdf)

    def sanity_check(self, silent: bool = True) -> str:
        """Verify BSDF data is correctly defined.

        Parameters
        ----------
        silent : bool
            If False Warnings will be raised else not, by Default True

        Returns
        -------
        WarningInformation : str
            Description of what data is missing or incorrect
        """
        return self._sanity_check(raise_error=False, silent=silent)

    def _sanity_check(self, raise_error=False, silent=True):
        """Validate data.

        Allow to raise an error
        """
        r_wl = []
        r_inc = []
        t_wl = []
        t_inc = []
        error_msg = ""
        match self.has_reflection, self.has_transmission:
            case True, True:
                for brdf, btdf in zip(self.brdf, self.btdf):
                    r_inc.append(brdf.incident_angle)
                    r_wl.append(brdf.wavelength)
                    t_inc.append(btdf.incident_angle)
                    t_wl.append(btdf.wavelength)
                if r_inc != t_inc or r_wl != t_wl:
                    error_msg += (
                        "Incidence and/or Wavelength information between reflection and"
                        " transmission is not identical. "
                    )
                test_inc = r_inc
                test_wl = r_wl
            case True, False:
                for brdf in self.brdf:
                    r_inc.append(brdf.incident_angle)
                    r_wl.append(brdf.wavelength)
                test_inc = r_inc
                test_wl = r_wl
            case False, True:
                for btdf in self.btdf:
                    t_inc.append(btdf.incident_angle)
                    t_wl.append(btdf.wavelength)
                test_inc = t_inc
                test_wl = t_wl
            case _:
                test_inc = []
                test_wl = []
        inc_f = dict(Counter(test_inc))
        wl_f = dict(Counter(test_wl))
        inc_error_l = []
        wl_error_l = []
        for key in inc_f:
            if len(wl_f.keys()) != inc_f[key]:
                inc_error_l.append(key)
        for key in wl_f:
            if len(inc_f.keys()) != wl_f[key]:
                wl_error_l.append(key)
        if inc_error_l:
            error_msg += (
                "The bsdf is missing information's for the for the following incidence"
                " angles one or more wavelengths are missing: {}. ".format(inc_error_l)
            )
        if inc_error_l:
            error_msg += (
                "The bsdf is missing information's for the for the following wavelength"
                " one or more incidence angles are missing: {}. ".format(wl_error_l)
            )
        if raise_error:
            if error_msg:
                raise ValueError(error_msg)
        elif silent:
            return error_msg
        else:
            if error_msg:
                warnings.warn(error_msg, stacklevel=2)
            return error_msg

    def commit(self):
        """Sent Data to gRPC interface."""
        # set basic values
        self._sanity_check(raise_error=True)
        spectral_bsdf = spectral_bsdf__v1__pb2.SpectralBsdfData()
        spectral_bsdf.description = self.description
        wl = []
        inc = []
        match self.has_reflection, self.has_transmission:
            case True, True:
                for brdf, btdf in zip(self.brdf, self.btdf):
                    inc.append(brdf.incident_angle)
                    wl.append(brdf.wavelength)
                    iw = spectral_bsdf.wavelength_incidence_samples.add()
                    iw.reflection.integral = brdf.tis
                    iw.reflection.phi_samples[:] = list(brdf.phi_values)
                    iw.reflection.theta_samples[:] = list(brdf.theta_values)
                    iw.reflection.bsdf_cos_theta[:] = brdf.bxdf.flatten().tolist()
                    iw.transmission.integral = btdf.tis
                    iw.transmission.phi_samples[:] = list(btdf.phi_values)
                    iw.transmission.theta_samples[:] = list(btdf.theta_values)
                    iw.transmission.bsdf_cos_theta[:] = btdf.bxdf.flatten().tolist()
            case True, False:
                for brdf in self.brdf:
                    inc.append(brdf.incident_angle)
                    wl.append(brdf.wavelength)
                    iw = spectral_bsdf.wavelength_incidence_samples.add()
                    iw.reflection.integral = brdf.tis
                    iw.reflection.phi_samples[:] = list(brdf.phi_values)
                    iw.reflection.theta_samples[:] = list(brdf.theta_values)
                    iw.reflection.bsdf_cos_theta[:] = brdf.bxdf.flatten().tolist()
            case False, True:
                for btdf in self.btdf:
                    inc.append(btdf.incident_angle)
                    wl.append(btdf.wavelength)
                    iw = spectral_bsdf.wavelength_incidence_samples.add()
                    iw.transmission.integral = btdf.tis
                    iw.transmission.phi_samples[:] = list(btdf.phi_values)
                    iw.transmission.theta_samples[:] = list(btdf.theta_values)
                    iw.transmission.bsdf_cos_theta[:] = btdf.bxdf.flatten().tolist()
        inc = list(set(inc))
        wl = list(set(wl))
        inc.sort()
        wl.sort()
        spectral_bsdf.incidence_samples[:] = inc
        spectral_bsdf.wavelength_samples[:] = wl
        self._stub.Import(spectral_bsdf)
        self._grpcbsdf = spectral_bsdf
        if self._BaseBSDF__interpolation_settings is not None:
            self._stub.SetSpecularInterpolationEnhancementData(
                self._BaseBSDF__interpolation_settings._InterpolationEnhancement__cones_data
            )

    def save(self, file_path: Union[Path, str], commit: bool = True) -> Path:
        """Save a Speos anistropic bsdf.

        Parameters
        ----------
        file_path : Union[Path, str]
            Filepath to save bsdf
        commit : bool
            commit data before saving

        Returns
        -------
        Path
            File location
        """
        file_path = Path(file_path)
        file_name = spectral_bsdf__v1__pb2.FileName()
        if commit:
            self.commit()
        else:
            self._stub.Import(self._grpcbsdf)
        if file_path.suffix == ".brdf":
            file_name.file_name = str(file_path)
        else:
            file_name.file_name = str(file_path.parent / (file_path.name + ".brdf"))
        self._stub.Save(file_name)
        return Path(file_name.file_name)


class BxdfDatapoint:
    """Class to store a BxDF data point.

    Parameters
    ----------
    is_brdf : bool
        true for transmittive date, False for reflective
    incident_angle : float
        incident angle in radian
    theta_values : Collection[float]
        list of theta values for the bxdf data matrix, in radian
    phi_values : Collection[float]
        list of phi values for the bxdf data matrix, in radian
    bxdf : Collection[float]
        nested list of bxdf values in 1/sr
    anisotropy : float
        Anisotropy angle in radian
    wavelength : float
        Wavelength in nm
    """

    def __init__(
        self,
        is_brdf: bool,
        incident_angle: float,
        theta_values: Collection[float],
        phi_values: Collection[float],
        bxdf: Collection[float],
        tis: float = 1,
        anisotropy: float = 0,
        wavelength: float = 555,
    ):
        # data_reset
        self._theta_values = []
        self._phi_values = []
        self._bxdf = None
        # define data
        self.is_brdf = is_brdf
        self.incident_angle = incident_angle
        self.anisotropy = anisotropy
        self.theta_values = theta_values
        self.phi_values = phi_values
        self.bxdf = bxdf
        self.tis = tis
        self.wavelength = wavelength

    def get(self, key=""):
        """Retrieve any information from the BxdfDatapoint object.

        Parameters
        ----------
        key : str
            Name of the property.

        Returns
        -------
        property
            Values/content of the associated property.
        """
        data = {
            k: v.fget(self) for k, v in BxdfDatapoint.__dict__.items() if isinstance(v, property)
        }
        if key == "":
            return data
        elif data.get(key):
            return data.get(key)
        else:
            print("Used key: {} not found in key list: {}.".format(key, data.keys()))

    def __str__(self):
        """Create string representation of a RayPath."""
        return str(self.get())

    @property
    def anisotropy(self):
        """Anisotropy angels of Datapoint."""
        return self._anisotropy

    @anisotropy.setter
    def anisotropy(self, value):
        if 0 <= value <= 2 * np.pi:
            self._anisotropy = value
        else:
            raise ValueError("Anisotropy angle needs to be between [0, 2*pi]")

    @property
    def bxdf(self) -> np.array:
        """BxDF data as np matrix in 1/sr.

        Returns
        -------
        np.array:
            bxdf data in shape theta_values, phi_values

        """
        return self._bxdf

    @bxdf.setter
    def bxdf(self, value: Collection[float]):
        if value is not None:
            bxdf = np.array(value)
            if any((bxdf < 0).flatten()):
                raise ValueError("bxdf data has to be positive")
            if np.shape(bxdf) == (len(self.theta_values), len(self.phi_values)):
                self._bxdf = bxdf
            elif np.shape(bxdf) == (len(self.phi_values), len(self.theta_values)):
                self._bxdf = bxdf.transpose()
            else:
                raise ValueError("bxdf data has incorrect dimensions")
        else:
            self._bxdf = None

    @property
    def is_brdf(self):
        """Type of bxdf data point eitehr reflective or transmittive.

        Returns
        -------
        bool:
            true if reflective false if transmittive
        """
        return self._is_brdf

    @is_brdf.setter
    def is_brdf(self, value):
        self._is_brdf = bool(value)

    @property
    def incident_angle(self):
        """Incident angle of the Datapoint in radian.

        Returns
        -------
        float:
            Incidence angle in radian

        """
        return self._incident_angle

    @incident_angle.setter
    def incident_angle(self, value):
        if 0 <= value <= np.pi / 2:
            self._incident_angle = value
        else:
            raise ValueError("Incident angle needs to be between [0, pi/2]")

    def set_incident_angle(self, value, is_deg=True):
        """Allow to set an incident value in degree.

        Parameters
        ----------
        value : float
            value to be set
        is_deg : bool
            Allows to define if value is radian or degree
        """
        if is_deg:
            self.incident_angle = np.deg2rad(value)
        else:
            self.incident_angle = value

    @property
    def theta_values(self):
        """List of theta values for which values are stored in bxdf data."""
        return self._theta_values

    @theta_values.setter
    def theta_values(self, value):
        if not self.is_brdf:
            if all([np.pi / 2 <= theta <= np.pi for theta in value]):
                self._theta_values = value
                if np.shape(self.bxdf) != (len(self.theta_values), len(self.phi_values)):
                    self.bxdf = None
            else:
                raise ValueError("Theta values for Transmission need to be between [pi/2, pi]")
        else:
            if all([0 <= theta <= np.pi / 2 for theta in value]):
                self._theta_values = value
                if np.shape(self.bxdf) != (len(self.theta_values), len(self.phi_values)):
                    self.bxdf = None
            else:
                raise ValueError("Theta values for Reflection need to be between [0, pi/2]")

    @property
    def phi_values(self):
        """List of phi values  for which values are stored in bxdf data."""
        return self._phi_values

    @phi_values.setter
    def phi_values(self, value):
        if all([0 <= phi <= 2 * np.pi for phi in value]):
            self._phi_values = value
            if np.shape(self.bxdf) != (len(self.theta_values), len(self.phi_values)):
                self.bxdf = None
        else:
            raise ValueError("Phi values need to be between [0, 2pi]")


def create_bsdf180(
    speos: ansys.speos.core.Speos,
    bsdf180_file_path: Union[str, Path],
    path_normal_bsdf: Union[str, Path],
    path_opposite_bsdf: Union[str, Path],
) -> Path:
    """Create a bsdf180 from 2 bsdf.

    This function allows to create BSDF180 from 2 bsdf files
    allowed files: *.coated *.brdf *.anisotropicbsdf *.scattering

    Parameters
    ----------
    speos : ansys.speos.core.Speos
        Speos Object to connect to RPC server
    bsdf180_file_path : Union[str, Path]
        File location of created bsdf180
    path_normal_bsdf : Union[str, Path]
        File location of first file, which represent normal direction
        Allowed files: *.coated, *.brdf, *.anisotropicbsdf, *.scattering
    path_opposite_bsdf : Union[str, Path]
        File location of first file, which represent anti-normal direction
        Allowed files: *.coated, *.brdf, *.anisotropicbsdf, *.scattering
    fix_disparity : bool
        This allows to create a bsdf when the two files are not normalized to each other.
        By default, ``False``

    Returns
    -------
    Path
        Returns where the file location of the bsdf180
    """
    supported = [".coated", ".brdf", ".anisotropicbsdf", ".scattering"]
    stub = bsdf_creation__v1__pb2_grpc.BsdfCreationServiceStub(speos.client.channel)
    bsdf180_file_path = Path(bsdf180_file_path)
    path_normal_bsdf = Path(path_normal_bsdf)
    path_opposite_bsdf = Path(path_opposite_bsdf)
    if path_normal_bsdf.suffix not in supported or path_opposite_bsdf.suffix not in supported:
        raise TypeError(
            f"Filetype not support please use one of the supported filetype, {supported}."
        )
    if bsdf180_file_path.suffix != ".bsdf180":
        bsdf180_file_path = bsdf180_file_path.parent / (bsdf180_file_path.name + ".bsdf180")
    bsdf180_request = bsdf_creation__v1__pb2.Bsdf180InputData()
    bsdf180_request.input_front_bsdf_file_name = str(path_normal_bsdf)
    bsdf180_request.input_opposite_bsdf_file_name = str(path_opposite_bsdf)
    bsdf180_request.output_file_name = str(bsdf180_file_path)
    stub.BuildBsdf180(bsdf180_request)
    return bsdf180_file_path


def create_spectral_brdf(
    speos: ansys.speos.core.Speos,
    spectral_bsdf_file_path: Union[str, Path],
    wavelength_list: list[float],
    anisotropic_bsdf_file_list: list[Union[Path, str]],
) -> Path:
    """Create a brdf from multiple bsdf.

    This function allows to create BRDF from multiple bsdf files
    allowed files: *.anisotropicbsdf

    Parameters
    ----------
    speos : ansys.speos.core.Speos
        Speos Object to connect to RPC server
    spectral_bsdf_file_path : Union[str, Path]
        File location of created BRDF file
    wavelength_list : list[float]
        List of wavelength
    anisotropic_bsdf_file_list :  list[Union[Path, str]]
        list of bsdf file locations

    Returns
    -------
    Path
        Location of created BRDF
    """
    stub = bsdf_creation__v1__pb2_grpc.BsdfCreationServiceStub(speos.client.channel)
    spectral_request = bsdf_creation__v1__pb2.SpectralBsdfInputData()
    spectral_bsdf_file_path = Path(spectral_bsdf_file_path)
    if spectral_bsdf_file_path.suffix != ".brdf":
        spectral_bsdf_file_path = spectral_bsdf_file_path.parent / (
            spectral_bsdf_file_path.name + ".brdf"
        )
    spectral_request.output_file_name = str(spectral_bsdf_file_path)
    if len(wavelength_list) == len(anisotropic_bsdf_file_list):
        anisotropic_bsdf_file_list = [Path(bsdf_loc) for bsdf_loc in anisotropic_bsdf_file_list]
        for bsdf_loc in anisotropic_bsdf_file_list:
            if bsdf_loc.suffix != ".anisotropicbsdf":
                raise TypeError("Filetype not support please use only anisotropicbsdf files.")
    else:
        raise RuntimeError("The Number BSDF file and wavelength needs to be identical")
    for wl, file in zip(wavelength_list, anisotropic_bsdf_file_list):
        tmp = spectral_request.input_anisotropic_samples.add()
        tmp.wavelength = float(wl)
        tmp.file_name = str(file)
    stub.BuildSpectralBsdf(spectral_request)
    return spectral_bsdf_file_path


def create_anisotropic_bsdf(
    speos: ansys.speos.core.Speos,
    anisotropic_bsdf_file_path: Union[str, Path],
    anisotropy_list: list[float],
    anisotropic_bsdf_file_list: list[Union[Path, str]],
    fix_disparity: bool = False,
) -> Path:
    """Create an anisotropic bsdf from anisotropic bsdf files.

    Parameters
    ----------
    speos : ansys.speos.core.Speos
        Speos Object to connect to RPC server
    anisotropic_bsdf_file_path : Union[str, Path]
        File location of created Anisotropic BSDF file
    anisotropy_list : list[float]
        ordered List of anisotropy value, in radian
    anisotropic_bsdf_file_list : list[Union[Path, str]]
        list of bsdf file locations
    fix_disparity : bool
        Fixes normalization disparity between BSDF,
        By default: ``False``

    Notes
    -----
    Please note that the bsdf files from the bsdf list need to be isotropic.

    Returns
    -------
    Path
        Location of created Anisotropic BSDF files
    """
    stub = bsdf_creation__v1__pb2_grpc.BsdfCreationServiceStub(speos.client.channel)
    ani_request = bsdf_creation__v1__pb2.AnisotropicBsdfInputData()
    anisotropic_bsdf_file_path = Path(anisotropic_bsdf_file_path)
    if anisotropic_bsdf_file_path.suffix != ".anisotropicbsdf":
        anisotropic_bsdf_file_path = anisotropic_bsdf_file_path.parent / (
            anisotropic_bsdf_file_path.name + ".anisotropicbsdf"
        )
    ani_request.output_file_name = str(anisotropic_bsdf_file_path)
    if len(anisotropy_list) == len(anisotropic_bsdf_file_list):
        anisotropic_bsdf_file_list = [Path(bsdf_loc) for bsdf_loc in anisotropic_bsdf_file_list]
        for bsdf_loc in anisotropic_bsdf_file_list:
            if bsdf_loc.suffix != ".anisotropicbsdf":
                raise TypeError("Filetype not support please use only anisotropicbsdf files.")
    else:
        raise RuntimeError("The Number BSDF file and wavelength needs to be identical")
    for ani, file in zip(anisotropy_list, anisotropic_bsdf_file_list):
        temp = ani_request.input_anisotropic_bsdf_samples.add()
        temp.anisotropic_angle = ani
        temp.file_name = str(file)
    ani_request.fix_disparity = fix_disparity
    ani_request.output_file_name = str(anisotropic_bsdf_file_path)
    stub.BuildAnisotropicBsdf(ani_request)
    return anisotropic_bsdf_file_path



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a wrapped abstraction of the gRPC proto API definition and stubs."""


class CrudStub:
    """Wraps a speos gRPC CRUD connection.

    This class is used as base class for all Speos databases interactions.
    Better use directly those inherited classes like SOPTemplateStub, SpectrumStub, ...
    """

    def __init__(self, stub):
        self._stubMngr = stub

    def create(self, request):
        """Create a new entry."""
        return self._stubMngr.Create(request)

    def read(self, request):
        """Get an existing entry."""
        return self._stubMngr.Read(request)

    def update(self, request):
        """Change an existing entry."""
        self._stubMngr.Update(request)

    def delete(self, request):
        """Remove an existing entry."""
        self._stubMngr.Delete(request)

    def list(self, request):
        """List existing entries."""
        return self._stubMngr.List(request)


class CrudItem:
    """Item of a database.

    Parameters
    ----------
    db : ansys.speos.core.kernel.crud.CrudStub
        Database to link to.
    key : str
        Key (also named guid) of the item in the database.
    """

    def __init__(self, db: CrudStub, key: str):
        self._stub = db
        self._key = key

    @property
    def stub(self) -> CrudStub:
        """The database."""
        return self._stub

    @property
    def key(self) -> str:
        """The guid in database."""
        return self._key



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a way to interact with feature: Body."""

from __future__ import annotations

import re
from typing import List, Mapping, Optional, Union

from ansys.speos.core import proto_message_utils
import ansys.speos.core.face as face
from ansys.speos.core.geo_ref import GeoRef
from ansys.speos.core.kernel.body import ProtoBody
from ansys.speos.core.kernel.client import SpeosClient
import ansys.speos.core.part as part


class Body:
    """Feature : Body.

    Parameters
    ----------
    speos_client : ansys.speos.core.kernel.client.SpeosClient
        The Speos instance client.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Mapping[str, str]
        Metadata of the feature.
        By default, ``{}``.
    parent_part : Union[ansys.speos.core.part.Part, ansys.speos.core.part.Part.SubPart], optional
        Feature containing this sub part.
        By default, ``None``.

    Attributes
    ----------
    body_link : ansys.speos.core.kernel.body.BodyLink
        Link object for the body in database.
    """

    def __init__(
        self,
        speos_client: SpeosClient,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        parent_part: Optional[Union[part.Part, part.Part.SubPart]] = None,
    ) -> None:
        self._speos_client = speos_client
        self._parent_part = parent_part
        self._name = name
        self.body_link = None
        """Link object for the body in database."""

        if metadata is None:
            metadata = {}

        # Create local Body
        self._body = ProtoBody(name=name, description=description, metadata=metadata)

        self._geom_features = []

    @property
    def geo_path(self) -> GeoRef:
        """Geometry path to be used within other speos objects."""
        geo_paths = [self._name]
        if isinstance(self._parent_part, part.Part.SubPart):
            geo_paths.insert(0, self._parent_part.geo_path.metadata["GeoPath"])
        return GeoRef.from_native_link("/".join(geo_paths))

    def create_face(
        self,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
    ) -> face.Face:
        """Create a face in this element.

        Parameters
        ----------
        name : str
            Name of the feature.
        description : str
            Description of the feature.
            By default, ``""``.
        metadata : Optional[Mapping[str, str]]
            Metadata of the feature.
            By default, ``{}``.

        Returns
        -------
        ansys.speos.core.face.Face
            Face feature.
        """
        if metadata is None:
            metadata = {}

        face_feat = face.Face(
            speos_client=self._speos_client,
            name=name,
            description=description,
            metadata=metadata,
            parent_body=self,
        )
        self._geom_features.append(face_feat)
        return face_feat

    def _to_dict(self) -> dict:
        out_dict = ""

        if self.body_link is None:
            out_dict = proto_message_utils._replace_guids(
                speos_client=self._speos_client, message=self._body
            )
        else:
            out_dict = proto_message_utils._replace_guids(
                speos_client=self._speos_client, message=self.body_link.get()
            )

        return out_dict

    def __str__(self) -> str:
        """Return the string representation of the body."""
        out_str = ""

        if self.body_link is None:
            out_str += "local: "

        out_str += proto_message_utils.dict_to_str(dict=self._to_dict())
        return out_str

    def commit(self) -> Body:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.body.Body
            Body feature.
        """
        # Commit faces contained in this body
        for g in self._geom_features:
            g.commit()

        # Save or Update the body (depending on if it was already saved before)
        if self.body_link is None:
            self.body_link = self._speos_client.bodies().create(message=self._body)
        elif self.body_link.get() != self._body:
            self.body_link.set(data=self._body)  # Only Update if data has changed

        # Update the parent part
        if self._parent_part is not None:
            if self.body_link.key not in self._parent_part._part.body_guids:
                self._parent_part._part.body_guids.append(self.body_link.key)
                if self._parent_part.part_link is not None:
                    self._parent_part.part_link.set(data=self._parent_part._part)

        return self

    def reset(self) -> Body:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.body.Body
            Body feature.
        """
        # Reset body
        if self.body_link is not None:
            self._body = self.body_link.get()

        return self

    def delete(self) -> Body:
        """Delete feature: delete data from the speos server database.

        Returns
        -------
        ansys.speos.core.body.Body
            Body feature.
        """
        # Retrieve all features to delete them
        while len(self._geom_features) > 0:
            self._geom_features[0].delete()

        if self.body_link is not None:
            # Update the parent part
            if self._parent_part is not None:
                if self.body_link.key in self._parent_part._part.body_guids:
                    self._parent_part._part.body_guids.remove(self.body_link.key)
                    if self._parent_part.part_link is not None:
                        self._parent_part.part_link.set(data=self._parent_part._part)

            # Delete the body
            self.body_link.delete()
            self.body_link = None

        if self in self._parent_part._geom_features:
            self._parent_part._geom_features.remove(self)

        return self

    def find(
        self,
        name: str,
        name_regex: bool = False,
        feature_type: Optional[type] = None,
    ) -> List[face.Face]:
        """Find feature(s). In a body, only faces features can be found.

        Parameters
        ----------
        name : str
            Name of the feature.
            Example "FaceName"
        name_regex : bool
            Allows to use regex for name parameter.
            By default, ``False``, means that regex is not used for name parameter.
        feature_type : type
            Type of the wanted feature (example: ansys.speos.core.face.Face).
            By default, ``None``, means that all features will be considered.

        Returns
        -------
        List[ansys.speos.core.face.Face]
            Found features.
        """
        found_features = []
        if feature_type == face.Face or feature_type is None:
            if name_regex:
                p = re.compile(name)
                found_features.extend([x for x in self._geom_features if p.match(x._name)])
            else:
                found_features.extend([x for x in self._geom_features if x._name == name])

        return found_features



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module to start Speos RPC Server."""

import os
from pathlib import Path
import subprocess
import tempfile
from typing import Optional, Union

from ansys.speos.core import LOG as LOGGER
from ansys.speos.core.generic.constants import (
    DEFAULT_PORT,
    DEFAULT_VERSION,
    MAX_CLIENT_MESSAGE_SIZE,
    MAX_SERVER_MESSAGE_LENGTH,
)
from ansys.speos.core.generic.general_methods import retrieve_speos_install_dir
from ansys.speos.core.speos import Speos

try:
    import ansys.platform.instancemanagement as pypim

    _HAS_PIM = True
except ModuleNotFoundError:  # pragma: no cover
    _HAS_PIM = False


def launch_speos(version: str = None) -> Speos:
    """Start the Speos Service remotely using the product instance management API.

    Prerequisite : product instance management configured.

    Parameters
    ----------
    version : str, optional
        The Speos Service version to run, in the 3 digits format, such as "242".
        If unspecified, the version will be chosen by the server.

    Returns
    -------
    ansys.speos.core.speos.Speos
        An instance of the Speos Service.
    """
    if not _HAS_PIM:  # pragma: no cover
        raise ModuleNotFoundError(
            "The package 'ansys-platform-instancemanagement' is required to use this function."
        )

    if pypim.is_configured():
        LOGGER.info("Starting Speos service remotely. The startup configuration will be ignored.")
        return launch_remote_speos(version)


def launch_remote_speos(
    version: str = None,
) -> Speos:
    """Start the Speos Service remotely using the product instance management API.

    When calling this method, you need to ensure that you are in an
    environment where PyPIM is configured. This can be verified with
    :func:`pypim.is_configured <ansys.platform.instancemanagement.is_configured>`.

    Parameters
    ----------
    version : str, optional
        The Speos Service version to run, in the 3 digits format, such as "242".
        If unspecified, the version will be chosen by the server.

    Returns
    -------
    ansys.speos.core.speos.Speos
        An instance of the Speos Service.
    """
    if not _HAS_PIM:  # pragma: no cover
        raise ModuleNotFoundError(
            "The package 'ansys-platform-instancemanagement' is required to use this function."
        )

    pim = pypim.connect()
    instance = pim.create_instance(product_name="speos", product_version=version)
    instance.wait_for_ready()
    channel = instance.build_grpc_channel()
    return Speos(channel=channel, remote_instance=instance)


def launch_local_speos_rpc_server(
    version: str = DEFAULT_VERSION,
    port: Union[str, int] = DEFAULT_PORT,
    server_message_size: int = MAX_SERVER_MESSAGE_LENGTH,
    client_message_size: int = MAX_CLIENT_MESSAGE_SIZE,
    logfile_loc: str = None,
    log_level: int = 20,
    speos_rpc_path: Optional[Union[Path, str]] = None,
) -> Speos:
    """Launch Speos RPC server locally.

    .. warning::
        Do not execute this function with untrusted input parameters.

    Parameters
    ----------
    version : str
        The Speos server version to run, in the 3 digits format, such as "242".
        If unspecified, the version will be chosen as
        ``ansys.speos.core.kernel.client.LATEST_VERSION``.
    port : Union[str, int], optional
        Port number where the server is running.
        By default, ``ansys.speos.core.kernel.client.DEFAULT_PORT``.
    server_message_size : int
        Maximum message length value accepted by the Speos RPC server,
        By default, value stored in environment variable SPEOS_MAX_MESSAGE_LENGTH or 268 435 456.
    client_message_size: int
        Maximum Message size of a newly generated channel
        By default, ``MAX_CLIENT_MESSAGE_SIZE``.
    logfile_loc : str
        location for the logfile to be created in.
    log_level : int
        The logging level to be applied to the server, integer values can be taken from logging
        module.
        By default, ``logging.WARNING`` = 20.
    speos_rpc_path : Optional[str, Path]
        location of Speos rpc executable

    Returns
    -------
    ansys.speos.core.speos.Speos
        An instance of the Speos Service.
    """
    speos_rpc_path = retrieve_speos_install_dir(speos_rpc_path, version)
    if os.name == "nt":
        speos_exec = speos_rpc_path / "SpeosRPC_Server.exe"
    else:
        speos_exec = speos_rpc_path / "SpeosRPC_Server.x"
    if not logfile_loc:
        logfile_loc = Path(tempfile.gettempdir()) / ".ansys"
        logfile = logfile_loc / "speos_rpc.log"
    else:
        logfile = Path(logfile_loc)
        if logfile.is_file():
            logfile_loc = logfile.parent
        else:
            logfile_loc = Path(logfile_loc)
            logfile = logfile_loc / "speos_rpc.log"
    if not logfile_loc.exists():
        logfile_loc.mkdir()
    command = [
        str(speos_exec),
        "-p{}".format(port),
        "-m{}".format(server_message_size),
        "-l{}".format(str(logfile)),
    ]
    out, stdout_file = tempfile.mkstemp(suffix="speos_out.txt", dir=logfile_loc)
    err, stderr_file = tempfile.mkstemp(suffix="speos_err.txt", dir=logfile_loc)

    subprocess.Popen(command, stdout=out, stderr=err)
    return Speos(
        host="localhost",
        port=port,
        message_size=client_message_size,
        logging_level=log_level,
        logging_file=logfile,
        speos_install_path=speos_rpc_path,
    )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a way to interact with feature: Face."""

from __future__ import annotations

from typing import List, Mapping, Optional

from ansys.speos.core import proto_message_utils
import ansys.speos.core.body as body
from ansys.speos.core.geo_ref import GeoRef
from ansys.speos.core.kernel.client import SpeosClient
from ansys.speos.core.kernel.face import ProtoFace


class Face:
    """Feature : Face.

    Parameters
    ----------
    speos_client : ansys.speos.core.kernel.client.SpeosClient
        The Speos instance client.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    parent_body : ansys.speos.core.body.Body, optional
        Feature containing this face.
        By default, ``None``.

    Attributes
    ----------
    face_link : ansys.speos.core.kernel.face.FaceLink
        Link object for the face in database.
    """

    def __init__(
        self,
        speos_client: SpeosClient,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        parent_body: Optional[body.Body] = None,
    ) -> None:
        self._speos_client = speos_client
        self._parent_body = parent_body
        self._name = name
        self.face_link = None
        """Link object for the face in database."""
        if metadata is None:
            metadata = {}

        # Create local Face
        self._face = ProtoFace(name=name, description=description, metadata=metadata)

    @property
    def geo_path(self) -> GeoRef:
        """Geometry path to be used within other speos objects."""
        geo_paths = [self._name]
        parent = self._parent_body
        if isinstance(parent, body.Body):
            geo_paths.insert(0, parent.geo_path.metadata["GeoPath"])
        return GeoRef.from_native_link("/".join(geo_paths))

    def set_vertices(self, values: List[float]) -> Face:
        """Set the face vertices.

        Parameters
        ----------
        values : List[float]
            Coordinates of all points [p1x p1y p1z p2x p2y p2z ...].

        Returns
        -------
        ansys.speos.core.face.Face
            Face feature.
        """
        self._face.vertices[:] = values
        return self

    def set_facets(self, values: List[int]) -> Face:
        """Set the facets.

        Parameters
        ----------
        values : List[int]
            Indexes of points for all triangles (t1_1 t1_2 t1_3 t2_1 t2_2 t2_3 ...)

        Returns
        -------
        ansys.speos.core.face.Face
            Face feature.
        """
        self._face.facets[:] = values
        return self

    def set_normals(self, values: List[float]) -> Face:
        """Set the face normals.

        Parameters
        ----------
        values : List[float]
            Normal vectors for all points [n1x n1y n1z n2x n2y n2z ...]

        Returns
        -------
        ansys.speos.core.face.Face
            Face feature.
        """
        self._face.normals[:] = values
        return self

    def _to_dict(self) -> dict:
        out_dict = ""

        if self.face_link is None:
            out_dict = proto_message_utils._replace_guids(
                speos_client=self._speos_client, message=self._face
            )
        else:
            out_dict = proto_message_utils._replace_guids(
                speos_client=self._speos_client, message=self.face_link.get()
            )

        return out_dict

    def __str__(self) -> str:
        """Return the string representation of the face."""
        out_str = ""

        if self.face_link is None:
            out_str += "local: "

        out_str += proto_message_utils.dict_to_str(dict=self._to_dict())
        return out_str

    def commit(self) -> Face:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.face.Face
            Face feature.
        """
        # Save or Update the face (depending on if it was already saved before)
        if self.face_link is None:
            self.face_link = self._speos_client.faces().create(message=self._face)
        elif self.face_link.get() != self._face:
            self.face_link.set(data=self._face)  # Only Update if data has changed

        # Update the parent body
        if self._parent_body is not None:
            if self.face_link.key not in self._parent_body._body.face_guids:
                self._parent_body._body.face_guids.append(self.face_link.key)
                if self._parent_body.body_link is not None:
                    self._parent_body.body_link.set(data=self._parent_body._body)

        return self

    def reset(self) -> Face:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.face.Face
            Face feature.
        """
        # Reset face
        if self.face_link is not None:
            self._face = self.face_link.get()

        return self

    def delete(self) -> Face:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.face.Face
            Face feature.
        """
        if self.face_link is not None:
            # Update the parent body
            if self._parent_body is not None:
                if self.face_link.key in self._parent_body._body.face_guids:
                    self._parent_body._body.face_guids.remove(self.face_link.key)
                    if self._parent_body.body_link is not None:
                        self._parent_body.body_link.set(data=self._parent_body._body)

            # Delete the face
            self.face_link.delete()
            self.face_link = None

            if self in self._parent_body._geom_features:
                self._parent_body._geom_features.remove(self)

        return self



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a way to interact with Speos feature: Intensity."""

from __future__ import annotations

from typing import List, Mapping, Optional

from ansys.speos.core.generic.general_methods import deprecate_kwargs
from ansys.speos.core.geo_ref import GeoRef
from ansys.speos.core.kernel.client import SpeosClient
from ansys.speos.core.kernel.intensity_template import ProtoIntensityTemplate
from ansys.speos.core.kernel.proto_message_utils import protobuf_message_to_dict
from ansys.speos.core.kernel.scene import ProtoScene
from ansys.speos.core.proto_message_utils import dict_to_str


class Intensity:
    """Speos feature : Intensity.

    By default, a lambertian intensity is created (cos with N=1 and total_angle=180).

    Parameters
    ----------
    speos_client : ansys.speos.core.kernel.client.SpeosClient
        The Speos instance client.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    intensity_props_to_complete : \
    ansys.api.speos.scene.v2.scene_pb2.Scene.SourceInstance.IntensityProperties, optional
        Intensity properties to complete.
        By default, ``None``.
    key : str
        Creation from an IntensityTemplateLink key

    Attributes
    ----------
    intensity_template_link : ansys.speos.core.kernel.intensity_template.IntensityTemplateLink
        Link object for the intensity template in database.
    """

    class Library:
        """Intensity of type: Library.

        By default, orientation as axis system is chosen and no exit geometries.

        Parameters
        ----------
        library : ansys.api.speos.intensity.v1.IntensityTemplate.Library
            Library to complete.
        library_props : \
        ansys.api.speos.scene.v2.scene_pb2.Scene.SourceInstance.IntensityProperties.LibraryProperties
            Library properties to complete.
        default_values : bool
            Uses default values when True.
        """

        def __init__(
            self,
            library: ProtoIntensityTemplate.Library,
            library_props: ProtoScene.SourceInstance.IntensityProperties.LibraryProperties,
            default_values: bool = True,
        ) -> None:
            self._library = library
            self._library_props = library_props

            if default_values:
                # Default values
                self.set_orientation_axis_system()

        def set_intensity_file_uri(self, uri: str) -> Intensity.Library:
            """Set the intensity file.

            Parameters
            ----------
            uri : str
                uri of the intensity file IES (.ies), Eulumdat (.ldt), speos intensities (.xmp)

            Returns
            -------
            ansys.speos.core.intensity.Intensity.Library
                Intensity feature of type library.
            """
            self._library.intensity_file_uri = uri
            return self

        def set_orientation_axis_system(
            self, axis_system: Optional[List[float]] = None
        ) -> Intensity.Library:
            """Set the intensity orientation from an axis system.

            Parameters
            ----------
            axis_system : Optional[List[float]]
                Orientation of the intensity [Ox Oy Oz Xx Xy Xz Yx Yy Yz Zx Zy Zz]
                By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]``.

            Returns
            -------
            ansys.speos.core.intensity.Intensity.Library
                Library intensity.
            """
            if not axis_system:
                axis_system = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]
            self._library_props.axis_system.values[:] = axis_system
            return self

        def set_orientation_normal_to_surface(self) -> Intensity.Library:
            """Set the intensity orientation as normal to surface.

            Returns
            -------
            ansys.speos.core.intensity.Intensity.Library
                Library intensity.
            """
            self._library_props.normal_to_surface.SetInParent()
            return self

        def set_orientation_normal_to_uv_map(self) -> Intensity.Library:
            """Set the intensity orientation as normal to uv map.

            Returns
            -------
            ansys.speos.core.intensity.Intensity.Library
                Library intensity.
            """
            self._library_props.normal_to_uv_map.SetInParent()
            return self

        def set_exit_geometries(
            self, exit_geometries: Optional[List[GeoRef]] = None
        ) -> Intensity.Library:
            """Set the exit geometries.

            Parameters
            ----------
            exit_geometries : Optional[List[ansys.speos.core.geo_ref.GeoRef]]
                Exit geometries list.
                By default, ``[]``.

            Returns
            -------
            ansys.speos.core.intensity.Intensity.Library
                Library intensity.
            """
            if not exit_geometries:
                self._library_props.ClearField("exit_geometries")
            else:
                self._library_props.exit_geometries.geo_paths[:] = [
                    gr.to_native_link() for gr in exit_geometries
                ]
            return self

    class Gaussian:
        """Intensity of type: Gaussian.

        By default, full width at half maximum following x and y are set at 30 degrees, and total
        angle at 180 degrees.
        By default, no axis system is chosen, that means normal to surface map.

        Parameters
        ----------
        gaussian : ansys.api.speos.intensity.v1.IntensityTemplate.Gaussian
            Gaussian to complete.
        gaussian_props : \
        ansys.api.speos.scene.v2.scene_pb2.Scene.SourceInstance.IntensityProperties.GaussianProperties
            Gaussian properties to complete.
        default_values : bool
            Uses default values when True.
        """

        def __init__(
            self,
            gaussian: ProtoIntensityTemplate.Gaussian,
            gaussian_props: ProtoScene.SourceInstance.IntensityProperties.GaussianProperties,
            default_values: bool = True,
        ) -> None:
            self._gaussian = gaussian
            self._gaussian_props = gaussian_props

            if default_values:
                # Default values
                self.set_FWHM_angle_x().set_FWHM_angle_y().set_total_angle().set_axis_system()

        def set_FWHM_angle_x(self, value: float = 30) -> Intensity.Gaussian:
            """Set the full width following x at half maximum.

            Parameters
            ----------
            value : float
                Full Width in degrees following x at Half Maximum.
                By default, ``30.0``.

            Returns
            -------
            ansys.speos.core.intensity.Intensity.Gaussian
                Gaussian intensity.
            """
            self._gaussian.FWHM_angle_x = value
            return self

        def set_FWHM_angle_y(self, value: float = 30) -> Intensity.Gaussian:
            """Set the full width following y at half maximum.

            Parameters
            ----------
            value : float
                Full Width in degrees following y at Half Maximum.
                By default, ``30.0``.

            Returns
            -------
            ansys.speos.core.intensity.Intensity.Gaussian
                Gaussian intensity.
            """
            self._gaussian.FWHM_angle_y = value
            return self

        def set_total_angle(self, value: float = 180) -> Intensity.Gaussian:
            """Set the total angle of the emission of the light source.

            Parameters
            ----------
            value : float
                Total angle in degrees of the emission of the light source.
                By default, ``180.0``.

            Returns
            -------
            ansys.speos.core.intensity.Intensity.Gaussian
                Gaussian intensity.
            """
            self._gaussian.total_angle = value
            return self

        def set_axis_system(self, axis_system: Optional[List[float]] = None) -> Intensity.Gaussian:
            """Set the intensity distribution orientation.

            Parameters
            ----------
            axis_system : List[float], optional
                Orientation of the intensity distribution [Ox Oy Oz Xx Xy Xz Yx Yy Yz Zx Zy Zz].
                By default, ``None`` : normal to surface map.

            Returns
            -------
            ansys.speos.core.intensity.Intensity.Gaussian
                Gaussian intensity.
            """
            self._gaussian_props.Clear()
            if axis_system is None:
                self._gaussian_props.SetInParent()
            else:
                self._gaussian_props.axis_system[:] = axis_system
            return self

    def __init__(
        self,
        speos_client: SpeosClient,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        intensity_props_to_complete: Optional[ProtoScene.SourceInstance.IntensityProperties] = None,
        key: str = "",
    ) -> None:
        self._client = speos_client
        self.intensity_template_link = None
        """Link object for the intensity template in database."""

        if metadata is None:
            metadata = {}

        # Attribute representing the more complex intensity.
        self._type = None

        # Create IntensityProperties
        self._intensity_properties = ProtoScene.SourceInstance.IntensityProperties()
        self._light_print = False
        if intensity_props_to_complete is not None:
            self._intensity_properties = intensity_props_to_complete
            self._light_print = True

        if key == "":
            # Create IntensityTemplate
            self._intensity_template = ProtoIntensityTemplate(
                name=name, description=description, metadata=metadata
            )

            # Default values
            self.set_cos(n=1)  # By default will be lambertian (cos with N =1)
        else:
            # Retrieve IntensityTemplate
            self.intensity_template_link = speos_client[key]
            self._intensity_template = self.intensity_template_link.get()

    def set_library(self) -> Intensity.Library:
        """Set the intensity as library.

        Returns
        -------
        ansys.speos.core.intensity.Intensity.Library
            Library intensity.
        """
        if self._type is None and self._intensity_template.HasField("library"):
            # Happens in case of project created via load of speos file
            self._type = Intensity.Library(
                library=self._intensity_template.library,
                library_props=self._intensity_properties.library_properties,
                default_values=False,
            )
        elif not isinstance(self._type, Intensity.Library):
            # if the _type is not Library then we create a new type.
            self._type = Intensity.Library(
                library=self._intensity_template.library,
                library_props=self._intensity_properties.library_properties,
            )
        elif (
            self._type._library is not self._intensity_template.library
            or self._type._library_props is not self._intensity_properties.library_properties
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._type._library = self._intensity_template.library
            self._type._library_props = self._intensity_properties.library_properties
        return self._type

    @deprecate_kwargs({"N": "n"}, "0.3.0")
    def set_cos(self, n: float = 3, total_angle: float = 180) -> Intensity:
        """Set the intensity as cos.

        Parameters
        ----------
        n : float
            Order of cos law.
            By default, ``3``.
        total_angle : float
            Total angle in degrees of the emission of the light source.
            By default, ``180.0``.

        Returns
        -------
        ansys.speos.core.intensity.Intensity
            Intensity feature.
        """
        self._type = None
        self._intensity_template.cos.N = n
        self._intensity_template.cos.total_angle = total_angle
        self._intensity_properties.Clear()
        return self

    def set_gaussian(self) -> Intensity.Gaussian:
        """Set the intensity as gaussian.

        Returns
        -------
        ansys.speos.core.intensity.Intensity.Gaussian
            Gaussian intensity.
        """
        if self._type is None and self._intensity_template.HasField("gaussian"):
            # Happens in case of project created via load of speos file
            self._type = Intensity.Gaussian(
                gaussian=self._intensity_template.gaussian,
                gaussian_props=self._intensity_properties.gaussian_properties,
                default_values=False,
            )
        elif not isinstance(self._type, Intensity.Gaussian):
            # if the _type is not Gaussian then we create a new type.
            self._type = Intensity.Gaussian(
                gaussian=self._intensity_template.gaussian,
                gaussian_props=self._intensity_properties.gaussian_properties,
            )
        elif (
            self._type._gaussian is not self._intensity_template.gaussian
            or self._type._gaussian_props is not self._intensity_properties.gaussian_properties
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._type._gaussian = self._intensity_template.gaussian
            self._type._gaussian_props = self._intensity_properties.gaussian_properties
        return self._type

    @property
    def type(self) -> type:
        """Return type of sensor.

        Returns
        -------
        Example: None for lambertian or ansys.speos.core.intensity.Intensity.Library

        """
        return type(self._type)

    def _to_dict(self) -> dict:
        out_dict = {}
        if self.intensity_template_link is None:
            out_dict = protobuf_message_to_dict(self._intensity_template)
        else:
            out_dict = protobuf_message_to_dict(message=self.intensity_template_link.get())

        if self._light_print is False:
            out_dict["intensity_properties"] = protobuf_message_to_dict(
                message=self._intensity_properties
            )

        return out_dict

    def __str__(self) -> str:
        """Return the string representation of the intensity."""
        out_str = ""
        if self.intensity_template_link is None:
            out_str += "local: "
        out_str += dict_to_str(self._to_dict())
        return out_str

    def commit(self) -> Intensity:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.intensity.Intensity
            Intensity feature.
        """
        if self.intensity_template_link is None:
            self.intensity_template_link = self._client.intensity_templates().create(
                message=self._intensity_template
            )
        elif self.intensity_template_link.get() != self._intensity_template:
            self.intensity_template_link.set(
                data=self._intensity_template
            )  # Only update if template has changed

        return self

    def reset(self) -> Intensity:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.intensity.Intensity
            Intensity feature.
        """
        if self.intensity_template_link is not None:
            self._intensity_template = self.intensity_template_link.get()
        return self

    def delete(self) -> Intensity:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.intensity.Intensity
            Intensity feature.
        """
        if self.intensity_template_link is not None:
            self.intensity_template_link.delete()
            self.intensity_template_link = None
        return self



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a wrapped abstraction of the gRPC proto API definition and stubs."""

from typing import Iterator, List

from ansys.api.speos.part.v1 import (
    face_pb2 as messages,
    face_pb2_grpc as service,
)
from ansys.speos.core.kernel.crud import CrudItem, CrudStub
from ansys.speos.core.kernel.proto_message_utils import protobuf_message_to_str

ProtoFace = messages.Face
"""Face protobuf class : ansys.api.speos.part.v1.face_pb2.Face"""
ProtoFace.__str__ = lambda self: protobuf_message_to_str(self)


class FaceLink(CrudItem):
    """Link object for job in database.

    Parameters
    ----------
    db : ansys.speos.core.kernel.face.FaceStub
        Database to link to.
    key : str
        Key of the face in the database.
    """

    def __init__(self, db, key: str):
        super().__init__(db, key)

    def __str__(self) -> str:
        """Return the string representation of the face."""
        return str(self.get())

    def get(self) -> ProtoFace:
        """Get the datamodel from database.

        Returns
        -------
        face.Face
            Face datamodel.
        """
        return self._stub.read(self)

    def set(self, data: ProtoFace) -> None:
        """Change datamodel in database.

        Parameters
        ----------
        data : face.Face
            New Face datamodel.
        """
        self._stub.update(self, data)

    def delete(self) -> None:
        """Remove datamodel from database."""
        self._stub.delete(self)


class FaceStub(CrudStub):
    """
    Database interactions for face.

    Parameters
    ----------
    channel : grpc.Channel
        Channel to use for the stub.

    Examples
    --------
    The best way to get a FaceStub is to retrieve it from SpeosClient via faces() method.
    Like in the following example:

    >>> from ansys.speos.core.speos import Speos
    >>> speos = Speos(host="localhost", port=50098)
    >>> face_db = speos.client.faces()

    """

    def __init__(self, channel):
        super().__init__(stub=service.FacesManagerStub(channel=channel))
        self._actions_stub = service.FaceActionsStub(channel=channel)

    def create(self, message: ProtoFace) -> FaceLink:
        """Create a new entry.

        Parameters
        ----------
        message : face.Face
            Datamodel for the new entry.

        Returns
        -------
        ansys.speos.core.kernel.face.FaceLink
            Link object created.
        """
        resp = CrudStub.create(self, messages.Create_Request(face=ProtoFace(name="tmp")))

        chunk_iterator = FaceStub._face_to_chunks(
            guid=resp.guid, message=message, nb_items=128 * 1024
        )
        self._actions_stub.Upload(chunk_iterator)

        return FaceLink(self, resp.guid)

    def read(self, ref: FaceLink) -> ProtoFace:
        """Get an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.face.FaceLink
            Link object to read.

        Returns
        -------
        face.Face
            Datamodel of the entry.
        """
        if not ref.stub == self:
            raise ValueError("FaceLink is not on current database")
        chunks = self._actions_stub.Download(request=messages.Download_Request(guid=ref.key))
        return FaceStub._chunks_to_face(chunks)

    def update(self, ref: FaceLink, data: ProtoFace) -> None:
        """Change an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.face.FaceLink
            Link object to update.

        data : face.Face
            New datamodel for the entry.
        """
        if not ref.stub == self:
            raise ValueError("FaceLink is not on current database")

        CrudStub.update(
            self,
            messages.Update_Request(guid=ref.key, face=ProtoFace(name="tmp")),
        )
        chunk_iterator = FaceStub._face_to_chunks(guid=ref.key, message=data, nb_items=128 * 1024)
        self._actions_stub.Upload(chunk_iterator)

    def delete(self, ref: FaceLink) -> None:
        """Remove an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.face.FaceLink
            Link object to delete.
        """
        if not ref.stub == self:
            raise ValueError("FaceLink is not on current database")
        CrudStub.delete(self, messages.Delete_Request(guid=ref.key))

    def list(self) -> List[FaceLink]:
        """List existing entries.

        Returns
        -------
        List[ansys.speos.core.kernel.face.FaceLink]
            Link objects.
        """
        guids = CrudStub.list(self, messages.List_Request()).guids
        return list(map(lambda x: FaceLink(self, x), guids))

    @staticmethod
    def _face_to_chunks(guid: str, message: ProtoFace, nb_items: int) -> Iterator[messages.Chunk]:
        for j in range(4):
            if j == 0:
                chunk_face_header = messages.Chunk(
                    face_header=messages.Chunk.FaceHeader(
                        guid=guid,
                        name=message.name,
                        description=message.description,
                        metadata=message.metadata,
                        sizes=[
                            len(message.vertices),
                            len(message.facets),
                            len(message.texture_coordinates_channels),
                        ],
                    )
                )
                yield chunk_face_header
            elif j == 1:
                for i in range(0, len(message.vertices), nb_items):
                    chunk_vertices = messages.Chunk(
                        vertices=messages.Chunk.Vertices(data=message.vertices[i : i + nb_items])
                    )
                    yield chunk_vertices
            elif j == 2:
                for i in range(0, len(message.facets), nb_items):
                    chunk_facets = messages.Chunk(
                        facets=messages.Chunk.Facets(data=message.facets[i : i + nb_items])
                    )
                    yield chunk_facets
            elif j == 3:
                for i in range(0, len(message.normals), nb_items):
                    chunk_normals = messages.Chunk(
                        normals=messages.Chunk.Normals(data=message.normals[i : i + nb_items])
                    )
                    yield chunk_normals

    @staticmethod
    def _chunks_to_face(chunks: messages.Chunk) -> ProtoFace:
        out_face = ProtoFace()
        for chunk in chunks:
            if chunk.HasField("face_header"):
                out_face.name = chunk.face_header.name
                out_face.description = chunk.face_header.description
                out_face.metadata.update(chunk.face_header.metadata)
            if chunk.HasField("vertices"):
                out_face.vertices.extend(chunk.vertices.data)
            if chunk.HasField("facets"):
                out_face.facets.extend(chunk.facets.data)
            if chunk.HasField("normals"):
                out_face.normals.extend(chunk.normals.data)

        return out_face
